class LineageExpressionId:
    
    def __init__(self, node = None):
        if node is not None:
            exprId = node.exprId()
            self._id = exprId.id()
            self._name = node.name()
            self._sql = node.sql()
    
    # def __init__(self, id: int, name: str):
    #     self.__id = id
    #     self.__name = name
    #     self.__sql = None

    @property
    def id(self) -> int:
        return self._id
    
    @id.setter
    def id(self, value: str):
        self._id = value

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def sql(self) -> str:
        return self._sql

    @sql.setter
    def sql(self, value: str):
        self._sql = value

    def __str__(self):
        return f"{self.name}#{self.id} '{self.sql}'"

    # Serialization to JSON
    def to_json_object(self):
        return {
            "id": self._id,
            "name": self._name,
            "sql": self._sql,
            "_class": self.__class__.__name__
        }

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        #data = json.loads(json_str)
        instance = cls()
        instance.id = data.get("id")
        instance.name = data.get("name")
        instance.sql = data.get("sql")
        return instance


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.LineageSerializationHelper import LineageSerializationHelper
class LineageDependency:
    
    def __init__(self, source: LineageExpressionId = None, target: LineageExpressionId = None):
        self.__source = source
        self.__target = target
        #print(f"DEPENDENCY {self.__source} >>> {self.__target}")

    @property
    def source(self) -> LineageExpressionId:
        return self.__source

    @source.setter
    def source(self, value: LineageExpressionId):
        self.__source = value

    @property
    def target(self) -> LineageExpressionId:
        return self.__target

    @target.setter
    def target(self, value: LineageExpressionId):
        self.__target = value


    def __str__(self):
        return f"{self.__source} >>> {self.__target}"
    
    # Serialization to JSON
    def to_json_object(self):
        return {
            "source": self.__source.to_json_object(),
            "target": self.__target.to_json_object() #,
            #"_class": self.__class__.__name__
        }

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        #data = json.loads(json_str)
        instance = cls()
        instance.source = LineageSerializationHelper.deserialize_class(data.get("source"))
        instance.target = LineageSerializationHelper.deserialize_class(data.get("target"))
        return instance


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.LineageTableColumnId import LineageTableColumnId
from bifabrik.utils.lineage.LineageSerializationHelper import LineageSerializationHelper

class DataFrameLineageOutputColumn(LineageExpressionId):
    
    def __init__(self, column: LineageExpressionId = None, dependencies: list[LineageExpressionId] = []):
        if column is not None:
            self._id = column.id
            self._name = column.name
            self._sql = column.sql
        
        self._dependencies = dependencies

    @property
    def dependencies(self) -> list[LineageExpressionId]:
        return self._dependencies

    @property
    def tableColumnDependencies(self) -> list[LineageTableColumnId]:
         col_deps = list(filter(lambda x: isinstance(x, LineageTableColumnId), self._dependencies))
         return col_deps

    def __str__(self):
        res = f"{self.name}#{self.id} '{self.sql}'" #super().__str__()
        for dep in self._dependencies:
            res = res + f'\n--- {dep}'
        return res

    # Serialization to JSON
    def to_json_object(self):
        base = super().to_json_object()
        base['dependencies'] = [item.to_json_object() for item in self._dependencies]
        return base

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        #data = json.loads(json_str)
        instance = cls()
        instance._id = data.get("id")
        instance._name = data.get("name")
        instance._sql = data.get("sql")
        instance._dependencies = []
        deps = data.get("dependencies")
        for dep in deps:
            instance.dependencies.append(LineageSerializationHelper.deserialize_class(dep))
        return instance


"""
File system utilities - glob pattern search, Fabric path normalization and more
"""

import notebookutils.mssparkutils.fs
import glob2
import regex
import sempy.fabric as spf
import logging
import os
import datetime
import time
from IPython.display import clear_output
import json
import pyspark.sql.session as pss
import requests

__mounts = None
__defaultWorkspaceId = spf.get_notebook_workspace_id()
__defaultWorkspaceRefName = 'default'
__defaultLakehouseId = None
__defaultLakehouseWorkspaceId = None
__defaultLakehouseRefName = 'default'
#lgr = lg.getLogger()


def normalizeFileApiPath(path: str):
    """Normalizes a file path to the form of "/lakehouse/default/Files/folder/..."
    """
    # used internally for logging to a mounted directory (int the notebook's default lakehouse)
    r = path
    if not r.startswith('/'):
        r = '/' + r

    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    if lp.startswith(lhpt):
        lp = lp[len(lhpt):]
        r = r[len(lhpt):]
    if lp.startswith(dfpt):
        lp = lp[len(dfpt):]
        r = r[len(dfpt):]
    if lp.startswith(fpt):
        lp = lp[len(fpt):]
        r = r[len(fpt):]
    
    r = '/lakehouse/default/Files' + r
    return r

def normalizeRelativeSparkPath(path: str):
    """Normalizes a file path to the form of "Files/folder/file.csv"
    """
    r = path
    if not r.startswith('/'):
        r = '/' + r
    
    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    
    lp = r.lower()
    if lp.startswith(lhpt):
        lp = lp[len(lhpt):]
        r = r[len(lhpt):]
    if lp.startswith(dfpt):
        lp = lp[len(dfpt):]
        r = r[len(dfpt):]
    if lp.startswith(fpt):
        lp = lp[len(fpt):]
        r = r[len(fpt):]
    
    r = 'Files' + r
    return r

def normalizeAbfsPath(path: str, lakehouseBasePath: str, inFiles = False, inTables = False):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Files/CsvFiles or .../Tables/Table1"
    """
    r = path

    if not r.startswith('/'):
        r = '/' + r
    
    lhpt = '/lakehouse'
    dfpt = '/default'
    fpt = '/files'
    lp = r.lower()
    
    lp = r.lower()
    if lp.startswith('abfss:'):
        return r
    if not r.startswith('/'):
        r = '/' + r
        lp = '/' + lp
    
    if inFiles and not (r.startswith('/Files')):
        r = '/Files' + r
    elif inTables and not (r.startswith('/Tables')):
        r = '/Tables' + r
    
    wPrefix = f'{lakehouseBasePath}{r}'
    return wPrefix


def normalizeAbfsFilePath(path: str, lakehouseBasePath: str):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Files/CsvFiles"
    You can use getLakehousePath() to get the lakehouse path as "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com"
    """
    return normalizeAbfsPath(path, lakehouseBasePath=lakehouseBasePath, inFiles = True)

def normalizeAbfsTablePath(path: str, lakehouseBasePath: str):
    """Normalizes a path to the form of "abfss://{workspaceid}@onelake.dfs.fabric.microsoft.com/{lakehouseid}/Tables/Table1"
    """
    return normalizeAbfsPath(path, lakehouseBasePath=lakehouseBasePath, inTables = True)

def filePatternSearch(path: str, lakehouse: str = None, workspace: str = None, useImplicitDefaultLakehousePath = False) -> list[str]:
    """Searches the Files/ directory of the current lakehouse
    using glob to match patterns. Returns the list of files as ABFS paths.

    Examples
    --------
    >>> bifabrik.utils.fsUtils.filePatternSearch("fld1/*/data/*.csv")
    ...     ["abfss://...@onelake.dfs.fabric.microsoft.com/.../Files/fld1/subf1/data/file11.csv", "abfss://...@onelake.dfs.fabric.microsoft.com/.../Files/fld1/subf2/data/file22.csv"]

    param useImplicitDefaultLakehousePath indicates whether '/lakehouse/default' should be used to refer to the default lakehouse
    """
    
    
    log = logging.getLogger('bifabrik')
    res = []
    #pathNorm = normalizeRelativeSparkPath(path)

    lhPath = getLakehousePath(lakehouse = lakehouse, workspace = workspace, useImplicitDefaultLakehousePath=useImplicitDefaultLakehousePath)
    pathNorm = normalizeAbfsFilePath(path, lhPath)
    pathNormTrimSuffix = pathNorm[len(lhPath) + len('/Files/'):] #pathNorm[len('Files/'):]
    pathPts = pathNormTrimSuffix.split("/")
    startPath = lhPath + '/Files'
    defaultLh = '/lakehouse/default/'
    isInDefaultLakehouse = False
    if startPath.startswith(defaultLh):
        startPath = startPath[len(defaultLh):]
        isInDefaultLakehouse = True
    searchLocations = [startPath]

    if len(pathPts) == 0:
        return res
    
    for i in range(len(pathPts)):
        pathPt = pathPts[i]
        if len(searchLocations) == 0:
            return res
        nextLevel = []
        for location in searchLocations:
            log.info(f'Searching location {location}')
            if not notebookutils.mssparkutils.fs.exists(location):
                warn = f'Location {location} does not exist'
                print(warn)
                log.warning(warn)
                continue
            subLocations = notebookutils.mssparkutils.fs.ls(location)
            subLocationNames = [fi.name for fi in subLocations]
            subLocationsFilteredT = glob2.fnmatch.filter(subLocationNames, pathPt, True, False, None)
            subLocationsFiltered = [x[0] for x in subLocationsFilteredT] 
            #print(subLocationsFiltered)
            subLocationsDict = {}
            for sbl in subLocations:
                subLocationsDict[sbl.name] = sbl
            #print(subLocationsDict)
            for slf in subLocationsFiltered:
                #print(slf)
                finfo = subLocationsDict[slf]
                slfPath = location + '/' + slf
                if len(pathPts) == i + 1:
                    if finfo.isFile:
                        if isInDefaultLakehouse:
                            res.append(f'{defaultLh}{slfPath}')
                        else:
                            res.append(slfPath)
                elif (finfo.isDir):
                    nextLevel.append(slfPath)
        if len(pathPts) == i + 1:
            return res
        else:
            searchLocations = nextLevel

def fileExists(path: str, lakehouse: str = None, workspace: str = None) -> bool:
    searchResults = filePatternSearch(path = path, lakehouse = lakehouse, workspace = workspace)
    if len(searchResults) > 0:
        return True
    return False

def getMounts():
    global __mounts
    if __mounts is None:
        __mounts = notebookutils.mssparkutils.fs.mounts()
        clear_output(wait=True)
    return __mounts

def getDefaultLakehouseAbfsPath() -> str:
    for mp in getMounts():
        if mp.mountPoint == '/default':
            # print(f"Default Lakehouse is: {mp.source}")
            return mp.source
        return None
    
def isGuid(uuid):
    UUID_PATTERN = regex.compile(r'^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$', regex.IGNORECASE)
    #print(uuid)
    #print(type(uuid))
    r = bool(UUID_PATTERN.match(uuid))
    return r

def getWorkspaceAndLakehouseIdFromMountSource(mountSource):
    m = regex.match(r"abfss://(.+)@onelake.dfs.fabric.microsoft.com/(.+)", mountSource)
    return { 'workspaceId': m.captures(1)[0], 'lakehouseId': m.captures(2)[0] }

class LakehouseMeta:
    lakehouseName: str = None
    lakehouseId: str = None
    workspaceName: str = None
    workspaceId: str = None
    basePath: str = None
    schemasEnabled: bool = False

    def serialize(self) -> str:
        res = {
            'lakehouseName': self.lakehouseName,
            'lakehouseId':  self.lakehouseId,
            'workspaceName': self.workspaceName,
            'workspaceId': self.workspaceId,
            'basePath': self.basePath,
            'schemasEnabled': self.schemasEnabled
        }
        j = json.dumps(res, indent=4)
        return j
    @staticmethod
    def deserialize(json_data):
        d = json.loads(json_data)
        l = LakehouseMeta()
        l.lakehouseName = d['lakehouseName']
        l.lakehouseId = d['lakehouseId']
        l.workspaceName = d['workspaceName']
        l.workspaceId = d['workspaceId']
        l.basePath = d['basePath']
        l.schemasEnabled = d['schemasEnabled']
        return l

class LakehouseMap():

    def __init__(self, workspaceId: str, workspaceName: str):
        self.__dictById = {}
        self.__dictByName = {}
        self.workspaceId = workspaceId
        self.workspaceName = workspaceName
    
    def addLakehouse(self, lh: LakehouseMeta):
        self.__dictById[lh.lakehouseId] = lh
        self.__dictByName[lh.lakehouseName.lower()] = lh

    def __getitem__(self, key) -> LakehouseMeta:
        if isGuid(key):
            if key in self.__dictById:
                return self.__dictById[key]
            return None
        if key.lower() in self.__dictByName:
            return self.__dictByName[key.lower()]
        return None

    def __contains__(self,key):
        if self.__getitem__(key) is not None:
            return True
        return False
    
    def __iter__(self):
        for k in self.__dictById:
            yield self.__dictById[k]
    
    def serialize(self) -> str:
        res = { 'workspaceId': self.workspaceId, 'workspaceName': self.workspaceName }
        lhmaps = []
        for l in list(self.__dictById.values()):
            lhmaps.append(l.serialize())
        res['lakehouses'] = lhmaps
        j = json.dumps(res, indent=4)
        return j
    
    @staticmethod
    def deserialize(json_data):
        lm = LakehouseMap(None, None)
        d = json.loads(json_data)
        lm.workspaceId = d['workspaceId']
        lm.workspaceName = d['workspaceName']
        for l in d['lakehouses']:
            l_deser = LakehouseMeta.deserialize(l)
            lm.addLakehouse(l_deser)
        return lm

class WorkspaceMap:

    def __init__(self):
        self.__dictById = {}
        self.__dictByName = {}
    
    def addWorkspace(self, lm: LakehouseMap):
        self.__dictById[lm.workspaceId] = lm
        self.__dictByName[lm.workspaceName.lower()] = lm

    def __getitem__(self, key) -> LakehouseMap:
        #print(key)
        #print(self.__dictByName)
        if isGuid(key):
            if key in self.__dictById:
                return self.__dictById[key]
            return None
        if key.lower() in self.__dictByName:
            return self.__dictByName[key.lower()]
        return None
    
    def __contains__(self,key):
        if self.__getitem__(key) is not None:
            return True
        return False
    
    def __iter__(self):
        for k in self.__dictById:
            yield self.__dictById[k]
    
    def serialize(self) -> str:
        res = []
        for l in list(self.__dictById.values()):
            res.append(l.serialize())
        j = json.dumps(res, indent=4)
        return j

    @staticmethod
    def deserialize(json_data):
        m = WorkspaceMap()
        d = json.loads(json_data)
        for l in d:
            l_deser = LakehouseMap.deserialize(l)
            m.addWorkspace(l_deser)
        return m

__workspaceMap = WorkspaceMap()

defaultMount = getDefaultLakehouseAbfsPath()
if defaultMount is not None:
    wsLh = getWorkspaceAndLakehouseIdFromMountSource(defaultMount)    
    __defaultLakehouseId = wsLh['lakehouseId']
    __defaultLakehouseWorkspaceId = wsLh['workspaceId']

def mapLakehouses(workspaceId: str, workspaceName: str):
    lm = LakehouseMap(workspaceId = workspaceId, workspaceName = workspaceName)
    #__workspaceMap.addWorkspace(lm)
    #print(workspaceId)
    #print(workspaceName)
    #print('----')
    
    #lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)

    try:
        lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)
    except Exception as ex:
        # the lakehouse list API will start returning 429 when too many requests are received
        print(f'Waiting 70s - lakehouse list API congestion [{workspaceName}]')
        time.sleep(70)
        try:
            lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)
        except Exception as ex:
            print(f'Waiting 140s - lakehouse list API congestion [{workspaceName}]')
            time.sleep(140)
            try:
                lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)
            except Exception as ex:
                print(f'Waiting 280s - lakehouse list API congestion [{workspaceName}]')
                time.sleep(280)
                lhs = notebookutils.mssparkutils.lakehouse.list(workspaceId)

    for lh in lhs:
        try:
            l = LakehouseMeta()
            l.lakehouseName = lh.displayName
            l.lakehouseId = lh.id
            l.workspaceName = workspaceName
            l.workspaceId = workspaceId
            l.basePath = f"abfss://{workspaceId}@onelake.dfs.fabric.microsoft.com/{lh.id}"

            dbo_path = f"{l.basePath}/Tables/dbo"
            if notebookutils.mssparkutils.fs.exists(dbo_path):
                l.schemasEnabled = True
            else:
                l.schemasEnabled = False

            lm.addLakehouse(l)
        except Exception as ex:
            print(f'Error when loading lakehouses from {workspaceName}: {ex}')
    return lm

def mapWorkspacesUsingSempy():
    global __workspaceMap
    global __defaultWorkspaceId
    global __defaultWorkspaceRefName
    global __defaultLakehouseId
    global __defaultLakehouseRefName

    #wss = spf.list_workspaces()
    #display(wss)
    #print(type(wss[wss.Type == 'Workspace']))
    
    token = notebookutils.mssparkutils.credentials.getToken("https://api.fabric.microsoft.com/")
    headers = {"Authorization": f"Bearer {token}"}
    resp = requests.get("https://api.fabric.microsoft.com/v1/workspaces", headers=headers)
    wss = resp.json()['value']
    for ws in wss:
        if ws['type'] != 'Workspace':
            continue
        ws_id = ws['id']
        ws_name = ws['displayName']
        lm = mapLakehouses(workspaceId = ws_id, workspaceName = ws_name)
        __workspaceMap.addWorkspace(lm)
    #     print(ws_id)
    #     print(ws_name)
    # for wix, ws in wss[wss.Type == 'Workspace'].iterrows():
        # lm = mapLakehouses(workspaceId = ws['Id'], workspaceName = ws['Name'])
        # __workspaceMap.addWorkspace(lm)

    if __defaultWorkspaceId not in __workspaceMap:
        mwslm = mapLakehouses(workspaceId = __defaultWorkspaceId, workspaceName = 'My workspace')
        __workspaceMap.addWorkspace(mwslm)
    
    return __workspaceMap


def mapWorkspaces():
    global __workspaceMap

    spark = pss.SparkSession.builder.getOrCreate()
    map_config_key = 'bifabrik.fs.workspaceMap'
    
    if spark.sparkContext._conf.contains(map_config_key):
        print(f'reading lakehouse mapping from sparkContext._conf [{map_config_key}]')
        map_serialized = spark.sparkContext._conf.get(map_config_key)
        deser = WorkspaceMap.deserialize(map_serialized)
        __workspaceMap = deser
    else:
        mapWorkspacesUsingSempy()    
        #print(f'setting {map_config_key}')
        serialized = __workspaceMap.serialize()
        spark.sparkContext._conf.set(map_config_key, serialized)


mapWorkspaces()


def getLakehouseMeta(lakehouse: str, workspace: str = None, suppressNotFound = False) -> LakehouseMeta:
    """
    Finds the lakehouse amoung available workspaces and returns its metadata
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Files
    
    :param lakehouse: the lakehouse name or ID (if None, the current notebook's deault lakehouse is used)
    :param workspace: the name or ID of the workspace containg the lakehouse (if None, the workspace of the current notebook is used)
    
    :return: LakehouseMeta
    """
    global __workspaceMap
    global __defaultWorkspaceId
    global __defaultLakehouseId

    global __defaultWorkspaceRefName
    global __defaultLakehouseRefName

    # use the workspace of the notebook
    if workspace is None:
        workspace = __defaultWorkspaceId
    if workspace == __defaultLakehouseRefName:
        workspace = __defaultWorkspaceId

    # use the notebook's default lakehouse
    if lakehouse is None:
        lakehouse = __defaultLakehouseId
    if lakehouse == __defaultLakehouseRefName:
        lakehouse = __defaultLakehouseId

    errMsg = f'Could not find lakehouse {lakehouse} in workspace {workspace}'

    if lakehouse is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    
    lhMap = __workspaceMap[workspace]
    if lhMap is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    
    lh = lhMap[lakehouse]
    if lh is None:
        if suppressNotFound:
            return None
        raise Exception(errMsg)
    return lh

def currentLakehouse() -> LakehouseMeta:
    '''Info on the notebook's default (attached) lakehouse
    '''
    l =  getLakehouseMeta(lakehouse = __defaultLakehouseId, workspace = __defaultLakehouseWorkspaceId, suppressNotFound = True)
    if l is None: 
        return None
    return {
        'lakehouseName' : l.lakehouseName,
        'lakehouseId' : l.lakehouseId,
        'workspaceName' : l.workspaceName,
        'workspaceId' : l.workspaceId,
        'basePath' : l.basePath,
        'schemasEnabled': l.schemasEnabled
    } 

def currentLakehouseName() -> str:
    '''Name of the notebook's default (attached) lakehouse
    '''
    clh = currentLakehouse()
    if clh is None:
        return None
    return clh['lakehouseName']

def getLakehousePath(lakehouse: str, workspace: str = None, suppressNotFound = False, useImplicitDefaultLakehousePath = False):
    """
    Returns the lakehouse path as abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}
    it can then be appended as e.g
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Tables
        abfss://{workspace id}@onelake.dfs.fabric.microsoft.com/{lakehouse id}/Files
    
    :param lakehouse: the lakehouse name or ID (if None, the current notebook's deault lakehouse is used)
    :param workspace: the name or ID of the workspace containg the lakehouse (if None, the workspace of the current notebook is used)
    """
    meta = getLakehouseMeta(lakehouse, workspace, suppressNotFound)
    if meta is None:
        if suppressNotFound:
            return None
        errMsg = f'Could not find lakehouse {lakehouse} in workspace {workspace}'
        raise Exception(errMsg)
    
    if meta.workspaceId == __defaultWorkspaceId and meta.lakehouseId == __defaultLakehouseId and useImplicitDefaultLakehousePath:
        return '/lakehouse/default'

    return meta.basePath

def archiveFiles(files, archiveFolder, filePattern, lakehouse = None, workspace = None):
    if len(files) == 0:
        return
    if archiveFolder is None:
        raise Exception(f'Cannot archive the file {files[0]}, as the archive folder has not been set')
    if filePattern is None:
        raise Exception(f'Cannot archive the file {files[0]}, as the archive file patter has not been set')
    
    # ABFS path to the archive folder
    lhPath = getLakehousePath(lakehouse = lakehouse, workspace = workspace)
    folderPathNorm = normalizeAbfsFilePath(archiveFolder, lhPath) #.removesuffix('/')
    if not notebookutils.mssparkutils.fs.exists(folderPathNorm):
        notebookutils.mssparkutils.fs.mkdirs(folderPathNorm)

    for file in files:
        log = logging.getLogger('bifabrik')

        fname = os.path.basename(file)
        rootName, extension = os.path.splitext(fname)

        ct = datetime.datetime.now()
        cts = ct.strftime("%Y_%m_%d_%H_%M_%S_%f")
        
        fullPattern = '{folder}/' + filePattern
        fullTargetPath = fullPattern.format(folder = folderPathNorm, filename = rootName, timestamp = cts, extension = extension)

        log.info(f'Archiving {file} -> {fullTargetPath}')
        notebookutils.mssparkutils.fs.mv(file, fullTargetPath)

def getWorkspaceId(workspace):
    if workspace is None:
        return __defaultWorkspaceId
    lhMap = __workspaceMap[workspace]
    if lhMap is None:
        return None
    return lhMap.workspaceId


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
import re

class LineageTableColumnId(LineageExpressionId):
    
    def __init__(self, node = None, tableIdentifier: str = None):
        super().__init__(node)
        self._tableIdentifier = tableIdentifier
        self.__splitTableIdentifier()
    
    @property
    def tableIdentifier(self) -> str:
        return self._tableIdentifier
    
    @tableIdentifier.setter
    def tableIdentifier(self, value: str):
        self._tableIdentifier = value
        self.__splitTableIdentifier()
    
    def __str__(self):
        return f"{self._tableIdentifier}.`{self.name}`#{self.id}"

    # Serialization to JSON
    def to_json_object(self):
        base = super().to_json_object()
        base['tableIdentifier'] = self._tableIdentifier
        return base

    def __splitTableIdentifier(self):
        self._tableIdentifierParts = []
        if self._tableIdentifier is None:
            return
        parts = re.findall(r'`([^`]*)`', self._tableIdentifier)
        self._tableIdentifierParts = parts
    
    @property
    def tableName(self) -> str:
        if len(self._tableIdentifierParts) > 0:
            return self._tableIdentifierParts[len(self._tableIdentifierParts) - 1]
        else:
            return None
    
    @property
    def dbName(self) -> str:
        if len(self._tableIdentifierParts) > 1:
            return self._tableIdentifierParts[len(self._tableIdentifierParts) - 2]
        else:
            return None

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        #data = json.loads(json_str)
        instance = cls()
        instance.id = data.get("id")
        instance.name = data.get("name")
        instance.sql = data.get("sql")
        instance.tableIdentifier = data.get("tableIdentifier")
        return instance


"""Data lineage module
Track and save data lineage information for dataframes saved to the lakehouse as tables.

    Example usage
    -------------

    Notebook 1 - eanble lineage tracking and load a table
    --------------------------------------------
    
    >>> import bifabrik as bif
    >>>
    >>> # lineage tracking is disabled by default
    >>> bif.config.lineage.lineageEnabled = True
    >>> bif.config.lineage.lineageFolder = 'Files/bifabrik_lineage'
    >>>
    >>> # after you have enabled lineage, just load a table as per usual
    >>> (
    >>>     bif.fromSql('SELECT User_ID, Email, Alias FROM Bronze.daktela_users')
    >>>        .toTable('Users_Test')
    >>>        .destinationLakehouse('Silver')
    >>>        .run()
    >>> )
    

    Notebook 2 - process the lineage files in the folder
    (each load of a lakehouse table creates a JSON file in the folder)
    --------------------------------------------
    
    >>> from bifabrik.utils.lineage.DataFrameLineage import DataFrameLineage
    >>> 
    >>> file_path = f'/lakehouse/default/Files/bifabrik_lineage/Silver_Users_Test_20250701_192332_381.json'
    >>> with open(file_path, "r") as file:
    >>>     file_content = file.read()
    >>>
    >>> df_lineage = DataFrameLineage.fromJson(file_content)
    >>>
    >>> # handle the lineage info as needed
    >>>
    >>> # check for parsing errors - this shouldn't happen too often though :)
    >>> if df_lineage.error is not None:
    >>>     print(df_lineage.error)
    >>> 
    >>> for c in df_lineage.columns:
    >>>     print(f'output column {c.name}')
    >>>     for dep in c.tableColumnDependencies:
    >>>         print(f'--- depends on {dep.dbName}.{dep.tableName}.{dep.name}')
    >>> 
    >>> # the context of the bifabrik load execution
    >>> print(df_lineage.context)
"""
import notebookutils.mssparkutils

import time
from datetime import datetime
from typing import List
import json
from typing import List
from typing import Set
from typing import Dict
import traceback

import bifabrik.utils.fsUtils as fsu

from bifabrik.utils.lineage.LineageContext import LineageContext
from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.DataFrameLineage import DataFrameLineage
from bifabrik.utils.lineage.DataFrameLineageOutputColumn import DataFrameLineageOutputColumn
from bifabrik.utils.lineage.LineageDependency import LineageDependency
from bifabrik.utils.lineage.LineageSerializationHelper import LineageSerializationHelper
from bifabrik.utils.lineage.VisitorPattern import visit_node

"""Save dataframe lineage information to the lakehouse as JSON."""
def save_df_lineage(df, targetTableName: str = None, targetLakehouseName: str = None, outputFolder = None):
    current_lakehouse = fsu.currentLakehouse()
    if current_lakehouse is None:
        print('Warning: No default lakehouse is not set, skipping data lineage')
        return
    
    #print('getting lineage information for dataframe')
    df_lineage = get_df_lineage(df, targetTableName, targetLakehouseName)
    if df_lineage is None:
        # this should never happen, but just in case
        return
    #print('OK')

    out_lakehouse_name = df_lineage.context.targetLakehouseName
    out_table_name = df_lineage.context.targetTableName
    if out_table_name is None:
        out_table_name = '__UNKNOWN_TABLE__'
    
    ts = df_lineage.context.executionDateTime
    ts_formatted = formatted = ts.strftime("%Y%m%d_%H%M%S") + f"_{ts.microsecond // 1000:03d}"
    file_name = f"{out_lakehouse_name}_{out_table_name}_{ts_formatted}.json"

    #print(f'Saving lineage information for {out_lakehouse_name}.{out_table_name} to {file_name}')
    logging_lh_path = current_lakehouse['basePath']

    targetFolderPath = fsu.normalizeAbfsFilePath(outputFolder, lakehouseBasePath = logging_lh_path)
    path = f'{outputFolder}/{file_name}'
    targetFilePath = fsu.normalizeAbfsFilePath(path, lakehouseBasePath = logging_lh_path)

    #print(f'Lineage file path: {targetFilePath}')
    notebookutils.mssparkutils.fs.mkdirs(targetFolderPath)
    lineage_json = df_lineage.to_json()
    notebookutils.mssparkutils.fs.put(targetFilePath, lineage_json, True)
    #print(f'Lineage information saved to {targetFilePath}')

"""Get dataframe lineage information as an object from the analyzed plan of the dataframe."""
def get_df_lineage(df, targetTableName: str = None, targetLakehouseName: str = None) -> DataFrameLineage:
    query_execution = df._jdf.queryExecution()
    analyzed_plan = query_execution.analyzed()

    analyzed_plan_str = str(analyzed_plan)
    target_lakehouse_name = targetLakehouseName
    target_table_name = targetTableName

    try:
        df_outputs = get_df_outputs(analyzed_plan)

        node_dependencies : list[LineageDependency] = []

        # for output in df_outputs:
        #     print(output)

        # apply the visitor pattern to the analyzed plan
        sub_references = visit_node(analyzed_plan, node_dependencies)

        # different nodes can have the same expression ID - LineageColumnId(s) (logical relation columns) take priority
        # afther that, longer SQL takes priority
        sub_references_ranked = sorted(sub_references, key=lambda x: (not(type(x).__name__ == "LineageTableColumnId"), -1 * len(x.sql)))

        # one expression ID can appear multiple times in the analyzed plan
        # only keep one instance - preferrably a table reference
        expressions_cleared: List[LineageExpressionId] = []
        expression_ids = set()

        # print('-----------')
        # print('EXPRESSIONS')
        # print('-----------')

        for expr in sub_references_ranked:
            #print(type(expr).__name__)
            #print(expr)
            if expr.id not in expression_ids:
                expressions_cleared.append(expr)
                expression_ids.add(expr.id)

        # for e in expressions_cleared:
        #     print(e)

        # print('------------')
        # print('DEPENDENCIES')
        # print('------------')

        dependencies_cleared = []
        dependencis_covered = set()

        # dependencies were discovered in the visitor pattern - aliases, expression sub-nodes etc.
        for dependency in node_dependencies:
            source_id = dependency.source.id
            target_id = dependency.target.id
            dep_tuple = (source_id, target_id)
            if dep_tuple in dependencis_covered:
                continue
            if source_id == target_id:
                continue
            dependencies_cleared.append(dependency)
            dependencis_covered.add(dep_tuple)


        # for dependency in dependencies_cleared:
        #     print(dependency)

        # find the transitive closure of dependencies for the output nodes
        # that is, all the dependencies leading to each output node

        # for each output, the list of its dependencies
        dependency_lists: Dict[int, List[LineageExpressionId]] = {}
        # dependencies grouped by destination (target) ID
        dependencies_by_destination: Dict[int, List[LineageDependency]] = {}

        for dependency in dependencies_cleared:
            if dependency.target.id not in dependencies_by_destination:
                dependencies_by_destination[dependency.target.id] = [dependency]
            else:
                dependencies_by_destination[dependency.target.id].append(dependency)

        expressions_cleared_by_id = {}
        for exp in expressions_cleared:
            expressions_cleared_by_id[exp.id] = exp

        # print('-------------------')
        # print('OUTPUT DEPENDENCIES')
        # print('-------------------')
    

        output_columns: list[DataFrameLineageOutputColumn] = []

        for output in df_outputs:
            #print(output)
            dependency_lists[output.id] = []
            covered_nodes = set()
            current_front_nodes = [output]
            while len(current_front_nodes) > 0:
                search_from_node = current_front_nodes.pop()
                covered_nodes.add(search_from_node.id)
                if search_from_node.id in dependencies_by_destination:
                    for dep in dependencies_by_destination[search_from_node.id]: 
                        if dep.source.id not in covered_nodes:
                            primary_expression = expressions_cleared_by_id[dep.source.id]
                            dependency_lists[output.id].append(primary_expression)
                            current_front_nodes.append(dep.source)
                            #print(f'--- {primary_expression}')
            
            # in some cases, there is no dependency - the column goes straight from the table (SELECT C FROM TABLE)
            # for those direct outputs, add a dependency to get the primary logical relation reference (using the expression ranking)
            
            # however, do not preface this with the following condition
            ### if len(dependency_lists[output.id]) == 0:
            # wince this "direct dependency" can occur in a UNION as well, where cle column already has a source, but not the one from the first query of the UNION
            
            if output.id in expressions_cleared_by_id:
                primary_expression = expressions_cleared_by_id[output.id]
                dependency_lists[output.id].append(primary_expression)
                    #print(f'--- {primary_expression}')

            output_columns.append(DataFrameLineageOutputColumn(output, dependency_lists[output.id]))
        
        result = DataFrameLineage(columns = output_columns, dependencies = dependencies_cleared, expressions = expressions_cleared
        ,logicalPlan = analyzed_plan_str, targetLakehouseName = target_lakehouse_name, targetTableName = target_table_name)
    
        return result

    except Exception as ex:
        stack_trace = ''.join(traceback.format_tb(ex.__traceback__))
        msg = f'{str(ex)}'
        stack_trace = stack_trace.replace("'", "\\\'")
        msg = msg.replace("'", "\\\'")
        full_msg = f'{msg}\n\nSTACK TRACE\n{stack_trace}'

        result = DataFrameLineage(logicalPlan = analyzed_plan_str, targetLakehouseName = target_lakehouse_name, targetTableName = target_table_name, error = full_msg)
        
        return result

def get_df_outputs(analyzed_plan) -> List[LineageExpressionId]:
    output_seq = analyzed_plan.output()
    output_list = [output_seq.apply(i) for i in range(output_seq.size())]
    res = list()
    for o in output_list:
        output_obj = LineageExpressionId(o)
        res.append(output_obj)
    return res


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class LineageConfiguration(Configuration):
    """Configuration related to saving data lineage information about dataframes saved to the lakehouse as tables
    """
    def __init__(self):
        super().__init__()
        self.__lineageEnabled = False
        self.__lineageFolder = 'Files/bifabrik_lineage'
        
    @CfgProperty
    def lineageEnabled(self) -> bool:
        """Whether to save lineage information about tables saved to the lakehouse
        default False
        """
        return self.__lineageEnabled
    @lineageEnabled.setter(key='lineageEnabled')
    def lineageEnabled(self, val):
        self.__lineageEnabled = val

    @CfgProperty
    def lineageFolder(self) -> bool:
        """Folder where lineage information about tables saved to the lakehouse is stored
        default Files/bifabrik_lineage
        """
        return self.__lineageFolder
    @lineageFolder.setter(key='lineageFolder')
    def lineageFolder(self, val):
        self.__lineageFolder = val



class Configuration:
    def __init__(self):
        self._explicitProps = {}

    def createCfgProperty(self, name, description):
        """Return a property that stores values under a private non-public name."""
        storage_name = '__' + name

        @CfgProperty
        def prop(self):
            return getattr(self, storage_name)

        @prop.setter(key=name)
        def prop(self, value):
            setattr(self, storage_name, value)
        
        prop.__doc__ = description
        
        setattr(self, name, prop)
        #return prop

class CfgProperty(property):
    def __set__(self, obj, value):
        super().__set__(obj, value)
        att = getattr(obj, '_explicitProps')
        att[self.key] = value

    def _setter(self, fset):
        obj = super().setter(fset)
        obj.key = self.key
        return obj

    def setter(self, key):
        self.key = key
        return self._setter


from datetime import datetime
import time
import json
import notebookutils.mssparkutils

class LineageContext:
    """Metadata about the notebook and bifabrik pipeline execution context"""
    def __init__(self, targetLakehouseName: str = None, targetTableName: str = None):
        self.__executionDateTime = time.time()
        self.__workspaceName = notebookutils.mssparkutils.runtime.context['currentWorkspaceName']
        self.__workspaceId = notebookutils.mssparkutils.runtime.context['currentWorkspaceId']
        self.__targetLakehouseName = notebookutils.mssparkutils.runtime.context['defaultLakehouseName'] if targetLakehouseName is None else targetLakehouseName
        self.__targetTableName = targetTableName
        self.__notebookName = notebookutils.mssparkutils.runtime.context['currentNotebookName']
        self.__notebookId = notebookutils.mssparkutils.runtime.context['currentNotebookId']
        self.__userName = notebookutils.mssparkutils.runtime.context['userName']
        self.__userId = notebookutils.mssparkutils.runtime.context['userId']

    @property
    def executionDateTime(self):
        """Execution time of the bifabrik table load"""
        edt = self.__executionDateTime
        return datetime.fromtimestamp(edt)

    @property
    def workspaceName(self):
        """Workspace in which the load was executed"""
        return self.__workspaceName

    @property
    def workspaceId(self):
        """Workspace in which the load was executed"""
        return self.__workspaceId

    @property
    def targetLakehouseName(self):
        """The target lakehouse to which bifabrik saved this data"""
        return self.__targetLakehouseName

    @property
    def targetTableName(self):
        """The target table to which bifabrik saved this data"""
        return self.__targetTableName

    @property
    def notebookName(self):
        """The name of the notebook in which bifabrik was executed"""
        return self.__notebookName

    @property
    def notebookId(self):
        """The ID of the notebook in which bifabrik was executed"""
        return self.__notebookId

    @property
    def userName(self):
        """The name of the user who executed the bifabrik table load"""
        return self.__userName

    @property
    def userId(self):
        """The ID of the user who executed the bifabrik table load"""
        return self.__userId

    # Serialization to JSON
    def to_json_object(self):
        return {
            "executionDateTime": self.__executionDateTime,
            "workspaceName": self.__workspaceName,
            "workspaceId": self.__workspaceId,
            "targetLakehouseName": self.__targetLakehouseName,
            "targetTableName": self.__targetTableName,
            "notebookName": self.__notebookName,
            "notebookId": self.__notebookId,
            "userName": self.__userName,
            "userId": self.__userId
        }
    
    def __str__(self):
        return json.dumps(self.to_json_object(), indent=4)

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        #data = json.loads(json_str)
        instance = cls(targetLakehouseName=data.get("targetLakehouseName"), targetTableName=data.get("targetTableName"))
        instance.__executionDateTime = data.get("executionDateTime", time.time())
        instance.__workspaceName = data.get("workspaceName")
        instance.__workspaceId = data.get("workspaceId")
        instance.__notebookName = data.get("notebookName")
        instance.__notebookId = data.get("notebookId")
        instance.__userName = data.get("userName")
        instance.__userId = data.get("userId")
        return instance


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.LineageTableColumnId import LineageTableColumnId

class LineageSerializationHelper:
    @classmethod
    def deserialize_class(cls, data):
        class_name = data["_class"]
        target_class = globals().get(class_name, cls)  # Lookup class by name

        result = getattr(target_class, 'from_json_object')(data)
        return result


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.LineageSerializationHelper import LineageSerializationHelper
from bifabrik.utils.lineage.DataFrameLineageOutputColumn import DataFrameLineageOutputColumn
from bifabrik.utils.lineage.LineageDependency import LineageDependency
from bifabrik.utils.lineage.LineageContext import LineageContext
import json

class DataFrameLineage:
    """
    The container for the lineage of a bifabrik pipeline saving data to a lakehouse table, 
    including its columns, dependencies, expressions, logical plan, and the notebook execution context.
    When lineage tracking is eanbled in the config, this object is created and saved as a JSON string to the bifabrik.config.lineage.lineageFolder folder in the default lakehouse.

    Example usage
    -------------
    
    >>> from bifabrik.utils.lineage.DataFrameLineage import DataFrameLineage
    >>> 
    >>> file_path = f'/lakehouse/default/Files/bifabrik_lineage/Silver_Users_Test_20250701_192332_381.json'
    >>> with open(file_path, "r") as file:
    >>>     file_content = file.read()
    >>>
    >>> df_lineage = DataFrameLineage.fromJson(file_content)
    >>>
    >>> # handle the lineage info as needed
    >>>
    >>> # check for parsing errors - this shouldn't happen too often though :)
    >>> if df_lineage.error is not None:
    >>>     print(df_lineage.error)
    >>> 
    >>> for c in df_lineage.columns:
    >>>     print(f'output column {c.name}')
    >>>     for dep in c.tableColumnDependencies:
    >>>         print(f'--- depends on {dep.dbName}.{dep.tableName}.{dep.name}')
    >>> 
    >>> # the context of the bifabrik load execution
    >>> print(df_lineage.context)
    """
    def __init__(self, columns: list[DataFrameLineageOutputColumn] = [], dependencies: list[LineageDependency] = [], expressions: list[LineageExpressionId] = [], 
        logicalPlan: str = None, targetLakehouseName: str = None, targetTableName: str = None, error = None):
        self.__columns = columns
        self.__dependencies = dependencies
        self.__expressions = expressions
        self.__logicalPlan = logicalPlan
        self.__error = error
        self.__context = LineageContext(targetLakehouseName = targetLakehouseName, targetTableName = targetTableName)
        

    # Getter and Setter for columns
    @property
    def columns(self) -> list["DataFrameLineageOutputColumn"]:
        return self.__columns

    @columns.setter
    def columns(self, value: list["DataFrameLineageOutputColumn"]):
        self.__columns = value

    # Getter and Setter for dependencies
    @property
    def dependencies(self) -> list["LineageDependency"]:
        return self.__dependencies

    @dependencies.setter
    def dependencies(self, value: list["LineageDependency"]):
        self.__dependencies = value

    # Getter and Setter for expressions
    @property
    def expressions(self) -> list["LineageExpressionId"]:
        return self.__expressions

    @expressions.setter
    def expressions(self, value: list["LineageExpressionId"]):
        self.__expressions = value

    # Getter and Setter for logicalPlan
    @property
    def logicalPlan(self) -> str:
        return self.__logicalPlan

    @logicalPlan.setter
    def logicalPlan(self, value: str):
        self.__logicalPlan = value

    # Getter and Setter for context
    @property
    def context(self) -> LineageContext:
        return self.__context

    @context.setter
    def context(self, value: LineageContext):
        self.__context = value

    @property
    def error(self) -> str:
        return self.__error

    @error.setter
    def error(self, value: str):
        self.__error = value

    def __str__(self):
        res = '\n'.join([str(item) for item in self.columns])
        return res
    
    def to_json_object(self):
        s_columns = []
        s_dependencies = []
        s_expressions = []

        for col in self.columns:
            s_columns.append(col.to_json_object())
        for dep in self.dependencies:
            s_dependencies.append(dep.to_json_object())
        for exp in self.expressions:
            s_expressions.append(exp.to_json_object())
        
        
        return {
            "columns": s_columns,
            "dependencies": s_dependencies,
            "expressions": s_expressions,
            "context": self.context.to_json_object(),
            "logicalPlan": self.logicalPlan,
            "error": self.error
        }

    # Deserialization from JSON
    @classmethod
    def from_json_object(cls, data):
        """Read a JSON-like object and deserialize a DataFrameLineage object."""
        #data = json.loads(json_str)
        # columns: list[DataFrameLineageOutputColumn], dependencies: list[LineageDependency], expressions: list[LineageExpressionId]
        instance = cls()
        instance.columns = [DataFrameLineageOutputColumn.from_json_object(item) for item in data['columns']]
        instance.dependencies = [LineageDependency.from_json_object(item) for item in data['dependencies']]
        instance.expressions = [LineageSerializationHelper.deserialize_class(item) for item in data['expressions']]
        instance.context = LineageContext.from_json_object(data['context'])
        instance.logicalPlan = data['logicalPlan']
        instance.error = data['error']
        
        return instance
    
    def to_json(self):
        """Serialize the DataFrameLineage object to a JSON string."""
        return json.dumps(self.to_json_object())

    def toJson(self):
        """Serialize the DataFrameLineage object to a JSON string."""
        return json.dumps(self.to_json_object())
    
    @classmethod
    def from_json(cls, json_string):
        """Read a JSON string and deserialize a DataFrameLineage object."""
        return cls.from_json_object(json.loads(json_string))
    
    @classmethod
    def fromJson(cls, json_string):
        """Read a JSON string and deserialize a DataFrameLineage object. This is the recommended method to use."""
        return cls.from_json_object(json.loads(json_string))


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.LineageConfiguration import LineageConfiguration

class LineageConfigurationContainer(ConfigContainer):

    def __init__(self):
        self.__lineage = LineageConfiguration()
        super().__init__()
        

    @property
    def lineage(self) -> LineageConfiguration:
        """Data lineage configuration"""
        return self.__lineage


from bifabrik.dst.DataDestination import DataDestination
from bifabrik.base.Pipeline import Pipeline
from bifabrik.cfg.TableDestinationConfiguration import TableDestinationConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils
import bifabrik.utils.log as lg
from bifabrik.utils import fsUtils
#from pyspark.sql.functions import col
from pyspark.sql.functions import *
from pyspark.sql.window import Window
import time
from datetime import datetime
import notebookutils.mssparkutils.fs
from bifabrik.utils import tableUtils as tu
import bifabrik.dst.CommonDestinationUtils as commons
from unidecode import unidecode
import bifabrik.utils.lineage as lineage_util

class LakehouseTableDestination(DataDestination, TableDestinationConfiguration):
    """Saves data to a lakehouse table.

    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromSql.query('SELECT * FROM SomeTable').toTable('DestinationTable1').run()
    """
    def __init__(self, pipeline: Pipeline, targetTableName: str, targetSchemaName: str = 'dbo'):
        super().__init__(pipeline)
        TableDestinationConfiguration.__init__(self)
        self.__targetTableName = targetTableName
        self.__targetSchemaName = targetSchemaName
        if '.' in targetTableName:
            split_name = targetTableName.split('.')
            self.__targetSchemaName = split_name[0]
            self.__targetTableName = split_name[1]

        self.__data = None
        self.__config = None
        self.__lhBasePath = None
        self.__identityColumn = None
        self.__insertDateColumn = None
        self.__updateDateColumn = None
        self.__tableExists = False
        self.__logger = None
        self.__tableLocation = None
        self.__addNARecord = False
        self.__addBadValueRecord = False
        self.__scd2RowStartTimestamp = None
        self.__scd2ExcludeColumns = []
        self.__partitionByColumns = []
        self.__partitionsDefined = False

    def __str__(self):
        return f'Table destination: {self.__targetTableName}'
    
    def execute(self, input: DataFrame) -> None:
        lgr = lg.getLogger()
        self.__logger = lgr
        self._error = None
        
        if input is None:
            msg = f'The table destination `{self.__targetTableName}` has no input; terminating'
            print(msg)
            lgr.warning(msg)
            return
        
        self.__data = input

        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__config = mergedConfig
        self.__tableConfig = mergedConfig.destinationTable
        
        # print(f'dst tbl cfg')
        # print(f'increment {self.__tableConfig.increment}')
        # print(f'mergeKeyColumns {self.__tableConfig.mergeKeyColumns}')
        # print(f'snapshotKeyColumns {self.__tableConfig.snapshotKeyColumns}')
        # print(f'identityColumnPattern {self.__tableConfig.identityColumnPattern}')
        # print(f'insertDateColumn {self.__tableConfig.insertDateColumn}')
        # print(f'invalidCharactersInColumnNamesReplacement {self.__tableConfig.invalidCharactersInColumnNamesReplacement}')

        dstLh = mergedConfig.destinationStorage.destinationLakehouse
        dstWs = mergedConfig.destinationStorage.destinationWorkspace
        self.__lhBasePath = fsUtils.getLakehousePath(dstLh, dstWs)
        self.__lhMeta = fsUtils.getLakehouseMeta(dstLh, dstWs)
        self.__useSchema = self.__lhMeta.schemasEnabled
        self.__sechemaRefPrefix = ''
        self.__tableLocation = self.__lhBasePath + "/Tables/" + self.__targetTableName
        if self.__useSchema:
            self.__sechemaRefPrefix = f'{self.__targetSchemaName}.'
            self.__tableLocation = self.__lhBasePath + "/Tables/" + self.__targetSchemaName + '/' + self.__targetTableName
        self.__addNARecord = self.__tableConfig.addNARecord
        self.__addBadValueRecord = self.__tableConfig.addBadValueRecord
        self.__scd2ExcludeColumns = self.__tableConfig.scd2ExcludeColumns
        self.__partitionByColumns = self.__tableConfig.partitionByColumns
        self.__partitionsDefined = len(self.__partitionByColumns) > 0
        self.__tableExists = self.__tableExistsF()

        incrementMethod = mergedConfig.destinationTable.increment
        if incrementMethod is None:
            incrementMethod = 'overwrite'
        incrementMethod = incrementMethod.lower()
        self.__incrementMethod = incrementMethod

        self.__replaceInvalidCharactersInColumnNames()
        
        identityColumnPattern = self.__tableConfig.identityColumnPattern
        if identityColumnPattern is not None:
            identityColumn = self.__tableConfig.identityColumnPattern.format(tablename = self.__targetTableName, lakehousename = self.__lhMeta.lakehouseName)
            self.__identityColumn = identityColumn

        # merge handles identity insert itself - separating the new records from the updates
        if (incrementMethod != 'merge' and incrementMethod != 'scd2') or self.__identityColumn is None or not(self.__tableExists):
            self.__insertIdentityColumn()

        #print('insert inserted date')
        self.__insertInsertDateColumn()
        self.__insertUpdateDateColumn()
        #print('insert N/A')
        self.__insertNARecord()
        #print('insert BadValue')
        self.__insertBadValueRecord()
        #print('insert scd2 tracking')
        self.__insertScd2TrackingColumns()

        #print('resolve schema differences')
        self.__resolveSchemaDifferences()
        
        self.__filterByWatermark()

        # if the target target table does not exist yet, just handle it as an overwrite
        if not(self.__tableExists):
            if incrementMethod == 'merge':
                # even if the table does not exist, check for duplicate merge keys
                self.__checkDuplicatesInMerge()
            self.__overwriteTarget()
        elif incrementMethod == 'overwrite':
            self.__overwriteTarget()
        elif incrementMethod == 'append':
            self.__appendTarget()
        elif incrementMethod == 'merge':
            self.__mergeTarget()
        elif incrementMethod == 'snapshot':
            self.__snapshotTarget()
        elif incrementMethod == 'scd2':
            self.__scd2Target()
        else:
            raise Exception(f'Unrecognized increment type: {incrementMethod}')

        if self.__tableConfig.ensureTableIsReadyAfterSaving:
            if not self.__data.rdd.isEmpty():
                table_name_full_check_existence = f"{self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}`{self.__targetTableName}`"
                self.__ensureTableIsReady(table_name_full_check_existence)

        self._result = self.__targetTableName

        #save data lineage
        lineage_config = mergedConfig.lineage
        if lineage_config is not None and lineage_config.lineageEnabled:
            self.__logger.info(f'Saving data lineage for table {self.__targetTableName} to {lineage_config.lineageFolder}')
            lineage_output_folder = lineage_config.lineageFolder
            lineage_util.save_df_lineage(input, targetTableName = self.__targetTableName, targetLakehouseName = self.__lhMeta.lakehouseName, outputFolder = lineage_output_folder)

        self._completed = True
        
    def __list_diff(self, first_list, second_list):
        diff = [item for item in first_list if item not in second_list]
        return diff
        
    def __replaceInvalidCharactersInColumnNames(self):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return
        
        self.__data = (
            self.
            __data
            .select(
                [col(f'`{c}`').alias(
                    self.__sanitizeColumnName(c)) 
                    for c in self.__data.columns
                ]))

    def __sanitizeColumnName(self, colName):
        replacement = self.__tableConfig.invalidCharactersInColumnNamesReplacement
        if replacement is None:
            return colName
        
        invalids = " ,;{}()\n\t="
        name = colName        
        for i in invalids:
            name = name.replace(i, replacement)
        name_without_accents = unidecode(name)
        return name_without_accents

    def __insertIdentityColumn(self):
        df_res = self.__insertIdentityColumnToDf(self.__data)
        self.__data = df_res

    def __insertIdentityColumnToDf(self, df):
        lgr = lg.getLogger()
        identityColumnPattern = self.__tableConfig.identityColumnPattern
        lgr.info(f'adding identity to df using pattern {identityColumnPattern}')
        if identityColumnPattern is None:
            return df
        
        lhName = self.__lhMeta.lakehouseName
        identityColumn = self.__tableConfig.identityColumnPattern.format(tablename = self.__targetTableName, lakehousename = lhName)
        self.__identityColumn = identityColumn

        initID = 0
        if self.__tableExists:
            query = f"SELECT MAX({identityColumn}) FROM {lhName}.{self.__sechemaRefPrefix}{self.__targetTableName}"
            initID = self._spark.sql(query).collect()[0][0]
            # if the table exists, but is empty
            if initID is None:
                initID = 0
        
        # place the identity column at the beginning of the table
        schema = [obj[0]  for obj in df.dtypes]
        schema.insert(0, identityColumn)
        df_ids  = df.withColumn(identityColumn, row_number().over(Window.orderBy(schema[1])).cast('bigint')).select(schema)
        df_res = df_ids.withColumn(identityColumn,col(identityColumn) + initID)
        
        return df_res

    def __insertInsertDateColumn(self):
        insertDateColumn = self.__tableConfig.insertDateColumn
        if insertDateColumn is None:
            return
        
        self.__insertDateColumn = insertDateColumn
        ts = time.time()
        r = self.__data.withColumn(insertDateColumn, lit(ts).cast("timestamp"))
        self.__data = r
    
    def __insertUpdateDateColumn(self):
        updateDateColumn = self.__tableConfig.updateDateColumn
        if updateDateColumn is None:
            return
        
        self.__updateDateColumn = updateDateColumn
        ts = time.time()
        r = self.__data.withColumn(updateDateColumn, lit(ts).cast("timestamp"))
        self.__data = r
    
    def __insertScd2TrackingColumns(self):        
        # self.__rowStartColumn = self.__tableConfig.rowStartColumn
        # row_end_column = self.__tableConfig.rowEndColumn
        # current_row_column = self.__tableConfig.currentRowColumn

        # src_filtered_sql = f'SELECT src.*, CURRENT_TIMESTAMP() AS {row_start_column}, CAST(9999-12-31 AS TIMESTMP) AS {row_end_column}, TRUE AS {current_row_column} \
        # FROM {src_view_name} src INNER JOIN {left_join_temp_name} AS lj ON {join_condition_src_left_join}'
        
        if self.__incrementMethod != 'scd2':
            return
        
        row_start_ts = ts = time.time()
        row_end = time.strptime("9999-12-31", '%Y-%m-%d')
        row_end_ts = time.mktime(row_end)
        self.__scd2RowStartTimestamp = row_start_ts
        r = self.__data \
            .withColumn(self.__tableConfig.rowStartColumn, lit(row_start_ts).cast("timestamp")) \
            .withColumn(self.__tableConfig.rowEndColumn, lit(row_end_ts).cast("timestamp")) \
            .withColumn(self.__tableConfig.currentRowColumn, lit(True))
        self.__data = r


    def __insertNARecord(self):
        if not self.__addNARecord:
            return
        # bad value and N/A records could cause duplicates in the merge columns
        # so insert these only when creating the table
        if self.__tableExists and self.__incrementMethod in ['merge', 'scd2']:
            return
        if self.__identityColumn is None:
            raise Exception('Configuration error - when addNARecord is enabled, identityColumnPattern needs to be configured as well.')
        if (not self.__tableExists) or (self.__incrementMethod in ['overwrite', 'snapshot']):
            self.__data = commons.addNARecord(self.__data, self._spark, self.__targetTableName, self.__identityColumn)
            return
        na_exists_sql = f'SELECT COUNT(*) FROM {self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}`{self.__targetTableName}` WHERE `{self.__identityColumn}` = -1'
        na_exists = self._spark.sql(na_exists_sql).collect()[0][0]
        if na_exists == 0:
            self.__data = commons.addNARecord(self.__data, self._spark, self.__targetTableName, self.__identityColumn)
    
    def __insertBadValueRecord(self):
        if not self.__addBadValueRecord:
            return
        # bad value and N/A records could cause duplicates in the merge columns
        # so insert these only when creating the table
        if self.__tableExists and self.__incrementMethod in ['merge', 'scd2']:
            return
        if self.__identityColumn is None:
            raise Exception('Configuration error - when addBadValueRecord is enabled, identityColumnPattern needs to be configured as well.')
        if (not self.__tableExists) or (self.__incrementMethod in ['overwrite', 'snapshot']):
            self.__data = commons.addBadValueRecord(self.__data, self._spark, self.__targetTableName, self.__identityColumn)
            return
        bv_exists_sql = f'SELECT COUNT(*) FROM {self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}`{self.__targetTableName}` WHERE `{self.__identityColumn}` = 0'
        bv_exists = self._spark.sql(bv_exists_sql).collect()[0][0]
        if bv_exists == 0:
            self.__data = commons.addBadValueRecord(self.__data, self._spark, self.__targetTableName, self.__identityColumn)

    def __overwriteTarget(self):
        writer = self.__data.write
        if self.__partitionsDefined:
            writer = writer.partitionBy(self.__partitionByColumns)

        writer.mode("overwrite").format("delta").option("overwriteSchema", "true").save(self.__tableLocation)

    def __appendTarget(self):
        writer = self.__data.write
        if self.__partitionsDefined:
            writer = writer.partitionBy(self.__partitionByColumns)

        writer.mode("append").format("delta").save(self.__tableLocation)

    def __resolveSchemaDifferences(self):
        lgr = lg.getLogger()

        lgr.info('resolve schema differences...')

        if not(self.__tableExists):
            return
        if self.__incrementMethod == 'overwrite':
            return
        
        target_table = f'{self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}{self.__targetTableName}'
        df_old = self._spark.sql(f"SELECT * FROM {target_table} LIMIT 0")
        cols_new = self._spark.createDataFrame(self.__data.dtypes, ["new_name", "new_type"])
        cols_old = self._spark.createDataFrame(df_old.dtypes, ["old_name", "old_type"])
        
        if self.__identityColumn is not None:
            cols_new = cols_new.filter(cols_new.new_name != self.__identityColumn)
            cols_old = cols_old.filter(cols_old.old_name != self.__identityColumn)

        compare = (
            cols_new.join(cols_old, cols_new.new_name == cols_old.old_name, how="outer")
            .fillna("")
        )
        # TODO: catch unsolvable type differences
        difference = compare.filter(compare.new_type != compare.old_type) \
            .filter((col('old_name') == '') | (col('new_name') == ''))

        canAddColumns = self.__tableConfig.canAddNewColumns
        allowMissingColumnsInSource = self.__tableConfig.allowMissingColumnsInSource

        
        # lgr.info('old table columns')
        # lgr.info(self.__getShowString(cols_old))
        # lgr.info('new table columns')
        # lgr.info(self.__getShowString(cols_new))

        # lgr.info('schema compare')
        # lgr.info(self.__getShowString(compare))
        lgr.info('schema difference')
        lgr.info(self.__getShowString(difference))
        #print('schema compare')
        #compare.show()
        #print('schema difference')
        #difference.show()


        # filter helper
        # dff = df.where( ~(~( ((df.LayerName == 'Gold') & (df.TableName.like('DM%'))) |  ((df.LayerName == 'Silver') & (df.TableName == 'Branches')) )) )
        # display(dff)

        # if canAddColumns:
        
        solvable = difference.where(((difference.old_name == '') & (canAddColumns == lit(True))) | ((difference.new_name == '') & (allowMissingColumnsInSource == lit(True))))
        insolvable = difference.where(~(((difference.old_name == '') & (canAddColumns == lit(True))) | ((difference.new_name == '') & (allowMissingColumnsInSource == lit(True)))))
        
            #insolvable = difference.where(difference.old_name != '')
        # else:
        #     solvable = difference.where(difference.old_type == '____none____')
        #     insolvable = difference

        if insolvable.count() > 0:
            err = f'Schema difference detected in table {target_table} that cannot be merged:'
            for r in insolvable.collect():
                ons = "N/A" if r.old_name == '' else r.old_name
                ots = "N/A" if r.old_type == '' else r.old_type
                nns = "N/A" if r.new_name == '' else r.new_name
                nts = "N/A" if r.new_type == '' else r.new_type
                err = err + f'\n> old_name: {ons}, old_type: {ots}'
                err = err + f', new_name: {nns}, new_type: {nts}'
            lgr.error(err)
            insolvable.show()
            raise Exception(err)

        for r in solvable.collect():
            
            # add new column to destination table
            if r.old_name == '':
                lgr.info(f'Adding new column {r.new_name} (r.new_type) to {self.__targetTableName}')
                tu.addTableColumnFromType(self.__lhMeta.lakehouseName, self.__targetTableName, r.new_name, r.new_type)
            # add missing column to source df
            elif r.new_name == '':
                lgr.info(f'Adding empty column {r.old_name} to source table')
                self.__data = self.__data.withColumn(r.old_name, lit(0))
            # this shouldnot really happen in the solvable set
            else:
                raise Exception(f'unexpected schema difference: {r}')



    def __mergeTarget(self):
        
        # print(str(datetime.now()))
        # print('running merge')

        lgr = lg.getLogger()
        all_columns = self.__data.columns
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.mergeKeyColumns))
        non_key_columns = self.__list_diff(self.__list_diff(all_columns, key_columns), [self.__identityColumn, self.__insertDateColumn])
        
        # print('key columns')
        # print(key_columns)
        # print('non-key columns')
        # print(non_key_columns)
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for merge increment. Please set the mergeKeyColumns property in destinationTable configuration to the list of column names.')
        

        # TODO: instead of SQL, use pyspark for cross-workspace ETL
        join_condition = " AND ".join([f"src.`{item}` = tgt.`{item}`" for item in key_columns])
        update_list = ", ".join([f"`{item}` = src.`{item}`" for item in non_key_columns])
        insert_list = ", ".join([f"`{item}`" for item in all_columns])
        insert_values = ", ".join([f"src.`{item}`" for item in all_columns])
        key_col_list = ", ".join([f"`{item}`" for item in key_columns])
        
        scd1_update = f"WHEN MATCHED THEN UPDATE SET \
            {update_list} "
        # table consisisting only of key columns - no need for the update clause
        if len(non_key_columns) == 0:
            scd1_update = ""

        src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        # temp table name in case we needed to use the large table approach
        src_temp_table_name_withid = f"temp_src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}_withid"
        

        merge_materialization_threshold = self.__tableConfig.mergeSourceMaterializationThresholdRowcount

        # if there are partitions, we're dealing with a big table
        # in that case, save the source to a temp table immediately to help with JOINs, instead of a view
        
        input_row_count = self.__data.count()
        if(input_row_count > merge_materialization_threshold):
                writer = self.__data.write
                if self.__partitionsDefined:
                    writer = writer.partitionBy(self.__partitionByColumns)
            
                # print(str(datetime.now()))
                # print(f'writing to initial src temp table {src_view_name}')

                writer.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{src_view_name}')
                
                self.__ensureTableIsReady(src_view_name)
                #time.sleep(10)
        else:
            self.__data.createOrReplaceTempView(src_view_name)
        
        mergeDbRef = f'{self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}' 

        # split data into news and updates
        
        # 1. separate new records dataframe

        src_news_df_sql = f'''
        SELECT /*+ BROADCAST(src) */ src.* 
        FROM {src_view_name} AS src
        LEFT JOIN {mergeDbRef}{self.__targetTableName} AS tgt ON {join_condition}
        WHERE tgt.`{key_columns[0]}` IS NULL
        '''
        # print('news df SQL')
        # print(src_news_df_sql)
        src_news_df = self._spark.sql(src_news_df_sql)
        
        # 2. generate IDs for the news DF independently - we don't want to mix these with the updates
        news_df = self.__insertIdentityColumnToDf(src_news_df)

        # 3. separate updated records dataframe 
        # and if identity column is configured, include in in the DF

        identityColumnPattern = self.__tableConfig.identityColumnPattern
        
        # print(str(datetime.now()))
        # print('handling identity')
        
        if identityColumnPattern is None:
            # SELECT /*+ BROADCAST(src) */ src.*
            src_updates_df = self._spark.sql(f'''
            SELECT src.* 
            FROM {src_view_name} AS src
            INNER JOIN {mergeDbRef}{self.__targetTableName} AS tgt ON {join_condition}
            ''')    
        else:
            insert_list = f'`{self.__identityColumn}`, {insert_list}'
            insert_values = f'src.`{self.__identityColumn}`, {insert_values}'
            src_updates_df_sql = f'''
            SELECT tgt.{self.__identityColumn}, src.* 
            FROM {src_view_name} AS src
            INNER JOIN {mergeDbRef}{self.__targetTableName} AS tgt ON {join_condition}
            '''

            # print('update df SQL')
            # print(src_updates_df_sql)
            src_updates_df = self._spark.sql(src_updates_df_sql)
        
        # 4. create an unified dataframe
        
        # print('news_df')
        # news_df.printSchema()
        # print('src_updates_df')
        # src_updates_df.printSchema()

        uni_df = news_df.union(src_updates_df)

        # get the target table size

        # print(str(datetime.now()))
        # print('getting the size of the target table')

        tableGB = 0
        inputGB = 0


        largeTableModeEnabled =  self.__tableConfig.largeTableMergeMethodEnabled
        sourceSizeThreshold = self.__tableConfig.largeTableMergeMethodSourceThresholdGB
        targetSizeThreshold = self.__tableConfig.largeTableMergeMethodDestinationThresholdGB

        targetTableGB = 0
        if largeTableModeEnabled:
            targetTableBytes = self._spark.sql(f'describe detail {mergeDbRef}{self.__targetTableName}').select("sizeInBytes").collect()[0][0]
            targetTableGB = targetTableBytes / 1024 / 1024 / 1024
            # empty placeholder value
            lgr.info(f'Target GB: {targetTableGB}')

        # print(str(datetime.now()))
        # print('size getting complete')

        # print(f'Large table mode enabled: {largeTableModeEnabled}')

        # to find out the source data sie, we need to save the source dataframe to a table
        # only do this if the target table size threshold is crossed
        # and the large table method is enabled

        if targetTableGB >= targetSizeThreshold and largeTableModeEnabled:

            lgr.info(f'writing to {src_temp_table_name_withid}')
            
            # print(str(datetime.now()))
            # print(f'writing to {src_temp_table_name_withid} - large table mode')

            #uni_df.write.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{src_temp_table_name_withid}')

            lgr.info(f'Partitioning source by {self.__partitionByColumns}')
            #print(f'Partitioning source by {self.__partitionByColumns}')
            writer = uni_df.write
            if self.__partitionsDefined:
                writer = writer.partitionBy(self.__partitionByColumns)
            writer.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{src_temp_table_name_withid}')
            
            # waiting for the new table to "come online"
            
            self.__ensureTableIsReady(src_temp_table_name_withid)
            #time.sleep(10)

            #uni_df.createOrReplaceTempView(src_temp_table_name_withid)
            #input_row_count = uni_df.count()
            lgr.info(f'input count {input_row_count}')
            try:
                inputBytes = self._spark.sql(f'describe detail {src_temp_table_name_withid}').select("sizeInBytes").collect()[0][0]
            except Exception as e:
                self.__logger.warning(f'failed to get the size of {src_temp_table_name_withid}')
                self.__logger.warning(e)
                time.sleep(20)
                inputBytes = self._spark.sql(f'describe detail {src_temp_table_name_withid}').select("sizeInBytes").collect()[0][0]
            inputGB = inputBytes / 1024 / 1024 / 1024
            lgr.info(f'Source GB: {inputGB}')
            # delete_src_temp_table_name_withid = True
        else:
            inputBytes = 0
            #uni_df.createOrReplaceTempView(src_temp_table_name_withid)

            # print(str(datetime.now()))
            # print('getting source count')
            # input_row_count = uni_df.count()

            # print(str(datetime.now()))
            # print(f'source count: {input_row_count}')

            if(input_row_count <= merge_materialization_threshold):
                uni_df.createOrReplaceTempView(src_temp_table_name_withid)
            else:
                
                # print(str(datetime.now()))
                # print(f'Partitioning source by {self.__partitionByColumns}')

                #print(f'Partitioning source by {self.__partitionByColumns}')
                writer = uni_df.write
                if self.__partitionsDefined:
                    lgr.info(f'Partitioning merge source by {self.__partitionByColumns}')
                    writer = writer.partitionBy(self.__partitionByColumns)
                writer.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{src_temp_table_name_withid}')
                

                # print(str(datetime.now()))
                # print(f'created {src_temp_table_name_withid}')
                
                # waiting for the new table to "come online"
                
                self.__ensureTableIsReady(src_temp_table_name_withid)
                #time.sleep(10)
                
        lgr.info('checking for duplicates')
        # print(str(datetime.now()))
        # print('checking for duplicates')
        duplicates_check = self._spark.sql(f'''
        SELECT {key_col_list}, COUNT(*) FROM {src_temp_table_name_withid}
        GROUP BY {key_col_list}
        HAVING COUNT(*) > 1
        ''')
        if duplicates_check.count() > 0:
            duplicates_check.show(10)
            raise Exception(f'There are duplicate records about to be merged into {self.__targetTableName}')

        #mergeDbRef = f'{self.__lhMeta.lakehouseName}.' 
        
        update_list = ", ".join([f"`{item}` = src.`{item}`" for item in non_key_columns])
        scd1_update = f"WHEN MATCHED THEN UPDATE SET \
            {update_list} "

        if tableGB < targetSizeThreshold or inputGB < sourceSizeThreshold or (not largeTableModeEnabled):
            # small table - direct update / insert
            merge_sql_update_basic = f"MERGE INTO {mergeDbRef}{self.__targetTableName} AS tgt \
            USING {src_temp_table_name_withid}  AS src \
            ON {join_condition} \
            {scd1_update}\
            "

            merge_sql_insert_basic = f"MERGE INTO {mergeDbRef}{self.__targetTableName} AS tgt \
            USING {src_temp_table_name_withid}  AS src \
            ON {join_condition} \
            WHEN NOT MATCHED THEN INSERT ( \
            {insert_list} \
            ) VALUES ( \
            {insert_values} \
            )\
            "

            lgr.info('merge update')
            lgr.info(merge_sql_update_basic)
            
            # print(str(datetime.now()))
            # print('merge update')
            # print(merge_sql_update_basic)
            
            self._spark.sql(merge_sql_update_basic)

            lgr.info('merge insert')
            lgr.info(merge_sql_insert_basic)
            
            # print(str(datetime.now()))
            # print('merge insert')
            # print(merge_sql_insert_basic)
            
            
            self._spark.sql(merge_sql_insert_basic)
            
            self._spark.sql(f'DROP TABLE IF EXISTS {src_view_name}')
            self._spark.sql(f'DROP TABLE IF EXISTS {src_temp_table_name_withid}')

            return
        
        msg = 'large table mrege strategy - rebuilding the target table'
        lgr.info(msg)
        print(msg)
        
        tgt_temp_table_name = f'temp_scd_tgt_{self.__targetTableName}'
        lgr.info(f'pooling rows for update to {tgt_temp_table_name}')
        tgt_df_query = f'''
        SELECT tgt.* 
        FROM {src_temp_table_name_withid} AS src
        INNER JOIN {mergeDbRef}{self.__targetTableName} AS tgt ON {join_condition}
        '''

        lgr.info(tgt_df_query)
        # print(tgt_df_query)
        tgt_df = self._spark.sql(tgt_df_query)

        tgt_df.write.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{tgt_temp_table_name}')

        # waiting for the new tables to "come online"
        
        self.__ensureTableIsReady(tgt_temp_table_name)
        #time.sleep(10)

        
        df_tgt_temp = self._spark.read.format('delta').load(f'Tables/{tgt_temp_table_name}')
        lgr.info(f'tgt temp count {df_tgt_temp.count()}')
        

        # 5. merge update and insert to the temp destination from source view
        
        merge_sql_update_temp = f"MERGE INTO {tgt_temp_table_name} AS tgt \
                    USING {src_temp_table_name_withid}  AS src \
                    ON {join_condition} \
                    {scd1_update}\
                    "
        lgr.info('merge update temp')
        lgr.info(merge_sql_update_temp)
        self._spark.sql(merge_sql_update_temp)

        merge_sql_insert_temp = f"MERGE INTO {tgt_temp_table_name} AS tgt \
            USING {src_temp_table_name_withid}  AS src \
            ON {join_condition} \
            WHEN NOT MATCHED THEN INSERT ( \
            {insert_list} \
            ) VALUES ( \
            {insert_values} \
            )\
            "
        lgr.info('merge insert temp')
        lgr.info(merge_sql_insert_temp)
        self._spark.sql(merge_sql_insert_temp)

        df_merged = self._spark.read.format('delta').load(f'Tables/{tgt_temp_table_name}')
        lgr.info(df_merged.count())

        # 5. add unmodified (unaffected, unmatched) rows from target to the temp table
        # (Xmerge delete from target table using temp destination (on merge keys))

        lgr.info('adding unchanged rows to temp')

        #tgt_temp_table_name = f'temp_scd_tgt_{self.__target_table_name}'
        tgt_unchanged_df_query = f'''
        SELECT tgt.* 
        FROM {mergeDbRef}{self.__targetTableName} AS tgt
        LEFT JOIN {src_temp_table_name_withid} AS src ON {join_condition}
        WHERE src.`{key_columns[0]}` IS NULL
        '''
        lgr.info(tgt_unchanged_df_query)
        tgt_unchanged_df = self._spark.sql(tgt_unchanged_df_query)

        lgr.info(tgt_unchanged_df.count())
        tgt_unchanged_df.write.mode("append").format("delta").save(f'Tables/{tgt_temp_table_name}')

        lgr.info('writing back to target table')

        #print('appending new and updated rows')
        #df_full = spark.sql(f'SELECT * FROM `{tgt_temp_table_name}`')
        df_full = self._spark.read.format('delta').load(f'Tables/{tgt_temp_table_name}')
        lgr.info(f'writing {df_full.count()} rows')
        
        writer = df_full.write
        if self.__partitionsDefined:
            writer = writer.partitionBy(self.__partitionByColumns)

        writer.write.mode("overwrite").format("delta").save(self.__tableLocation)

        self._spark.sql(f'DROP TABLE IF EXISTS {src_temp_table_name_withid}')
        self._spark.sql(f'DROP TABLE IF EXISTS {tgt_temp_table_name}')
        self._spark.sql(f'DROP TABLE IF EXISTS {src_view_name}')

    def __checkDuplicatesInMerge(self):
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.mergeKeyColumns))
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for merge increment. Please set the mergeKeyColumns property in destinationTable configuration to the list of column names.')

        duplicates = self.__data.groupBy(*key_columns).count().where(col('count') > 1)
        
        if duplicates.count() > 0:
            duplicates.show(10)
            raise Exception(f'There are duplicate records about to be merged into {self.__targetTableName}')


    def __scd2Target(self):
        lgr = lg.getLogger()
        row_start_column = self.__tableConfig.rowStartColumn
        row_end_column = self.__tableConfig.rowEndColumn
        current_row_column = self.__tableConfig.currentRowColumn
        all_columns = self.__data.columns
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.mergeKeyColumns))
        non_key_columns = self.__list_diff(self.__list_diff(all_columns, key_columns), [self.__identityColumn, self.__insertDateColumn, row_start_column, row_end_column, current_row_column])
        non_key_historized_columns = self.__list_diff(non_key_columns, self.__scd2ExcludeColumns)
        soft_deletes = self.__tableConfig.scd2SoftDelete
        
        #src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        # temp table name in case we needed to use the large table approach
        src_temp_table_name_withid = f"temp_src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}_withid"
        #self.__data.createOrReplaceTempView(src_view_name)
        db_reference = f'{self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}' 
        
        # print('key columns')
        # print(key_columns)
        # print('non-key columns')
        # print(non_key_columns)
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for SCD 2 increment. Please set the mergeKeyColumns property in destinationTable configuration to the list of column names.')
        
        # TODO: instead of SQL, use pyspark for cross-workspace ETL
        join_condition = " AND ".join([f"src.`{item}` = tgt.`{item}`" for item in key_columns])
        # update_list = ", ".join([f"`{item}` = src.`{item}`" for item in non_key_columns])
        # insert_list = ", ".join([f"`{item}`" for item in all_columns])
        # insert_values = ", ".join([f"src.`{item}`" for item in all_columns])
        key_col_list = ", ".join([f"`{item}`" for item in key_columns])

        # 1. check for duplicates
        # 2. if soft deletes are enabled, end-date deleted rows (source left join target...)
        # 3. target left join source to find matches
        # 4. for matched rows, find rows where all non-key attributes are the same as before and remove them from source
        # 5. end-date remaining matched current rows in target
        # 6. add RowStart, RowEnd and CurrentRow to source rows
        # 7. generate identity column for the source data if configured
        # 8. append all source rows to target


        # 1. check for duplicates

        deduplicated = self.__data.dropDuplicates(key_columns)
        orig_count = self.__data.count()
        deduplicated_count = deduplicated.count()
        if orig_count > deduplicated_count:
            #duplicates_df = self.__data.exceptAll(deduplicated)
            raise Exception(f'There are duplicate records about to be merged into {self.__targetTableName}')
        
        # 2. if soft deletes are enabled, end-date deleted rows (target left join source...)

        # will be used throughout the method
        src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        
        # materialize the source data to a temp table
        writer = self.__data.write
        if self.__partitionsDefined:
            writer = writer.partitionBy(self.__partitionByColumns)
        writer.mode("overwrite").format("delta").option("overwriteSchema", "true").save(f'Tables/{src_view_name}')
        #self.__data.createOrReplaceTempView(src_view_name)

        # waiting for the new table to "come online"
        self.__ensureTableIsReady(src_view_name)
        
        end_date_merge_update = f'SET tgt.{row_end_column} = CAST({self.__scd2RowStartTimestamp} AS TIMESTAMP), tgt.{current_row_column} = FALSE'
        #end_date_merge_update = f'UPDATE SET tgt.{row_end_column} = CURRENT_TIMESTAMP(), tgt.{current_row_column} = FALSE'
        
        if soft_deletes:
            soft_deletes_sql = f"""
            MERGE INTO {db_reference}{self.__targetTableName} AS tgt
            USING(
                SELECT {key_col_list}
                FROM {src_view_name}
                WHERE {current_row_column} = TRUE
            ) AS src
            ON {join_condition}
            WHEN NOT MATCHED BY SOURCE AND tgt.`{current_row_column}` = TRUE THEN UPDATE {end_date_merge_update}
            """
            
            lgr.info('running SQL merge to handle soft deletes')
            lgr.info('soft delete sql:')
            lgr.info(soft_deletes_sql)
            self._spark.sql(soft_deletes_sql)
        
        # 3. target left join source to find matches
        
# baded on https://medium.com/yipitdata-engineering/how-to-use-broadcasting-for-more-efficient-joins-in-spark-2d53a336b02b
# SELECT
# /*+ BROADCAST(small_df_a) */
# /*+ BROADCAST(small_df_b) */
# /*+ BROADCAST(small_df_c) */
# *
# FROM large_df
# LEFT JOIN small_df_a
# USING (id)
# LEFT JOIN small_df_b
# USING (id)
# LEFT JOIN small_df_c
# USING (id)

        #select_non_key_list = ", " + ", ".join([f"src.{item} AS src_{item}, tgt.{item} AS tgt_{item}" for item in non_key_columns])
        # non-key match condition


        key_match = " AND ".join([f"src.{item} <=> tgt.{item}" for item in key_columns])
        if len(non_key_historized_columns) == 0:
            combined_match = f"({key_match}) AS __scd2_combined_match"
        else:
            combined_match = f"({key_match} AND " + " AND ".join([f"src.{item} <=> tgt.{item}" for item in non_key_historized_columns]) + ") AS __scd2_combined_match"
        select_key_list = ", ".join([f"src.{item} AS src_{item}, tgt.{item} AS tgt_{item}" for item in key_columns])
        target_left_join_sql = f"""
SELECT
/*+ BROADCAST(src) */ 
{select_key_list}, {combined_match} 
FROM {src_view_name} AS src
LEFT JOIN {db_reference}{self.__targetTableName} AS tgt ON {join_condition} AND tgt.`{current_row_column}` = TRUE
"""
        lgr.info('running SQL left join to find matching rows')
        lgr.info('left join sql')
        lgr.info(target_left_join_sql)
        df_left_join = self._spark.sql(target_left_join_sql)
        left_join_temp_name = f"temp_match_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        
        # 4. for matched rows, find rows where all non-key attributes are the same as before and remove them from source

        #df_left_join.printSchema()
        df_left_join = df_left_join.filter('`__scd2_combined_match` = FALSE')
        df_left_join.createOrReplaceTempView(left_join_temp_name)
        self.__data.createOrReplaceTempView(src_view_name)
        
        #end_date_merge_update = f'SET tgt.{row_end_column} = CAST({self.__scd2RowStartTimestamp} AS TIMESTAMP), tgt.{current_row_column} = FALSE'
        #end_date_merge_update = f'SET tgt.{row_end_column} = CURRENT_TIMESTAMP(), tgt.{current_row_column} = FALSE'

        #src_view_name_filtered_with_id = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}_filtered_with_id"
        join_condition_src_left_join = " AND ".join([f"src.`{item}` = lj.`src_{item}`" for item in key_columns])
        
        # 6. add RowStart, RowEnd and CurrentRow to source rows
        
        src_filtered_sql = f'SELECT src.* \
        FROM {src_view_name} src INNER JOIN {left_join_temp_name} AS lj ON {join_condition_src_left_join}'

        #src_filtered_sql = f'SELECT src.*, CURRENT_TIMESTAMP() AS {row_start_column}, CAST(9999-12-31 AS TIMESTMP) AS {row_end_column}, TRUE AS {current_row_column} \
        #FROM {src_view_name} src INNER JOIN {left_join_temp_name} AS lj ON {join_condition_src_left_join}'

        lgr.info('filtered source query')
        lgr.info(src_filtered_sql)
        df_src_filtered = self._spark.sql(src_filtered_sql)
        
        # 7. generate identity column for the source data if configured

        df_src_filtered_with_id = self.__insertIdentityColumnToDf(df_src_filtered)
        
        #df_src_filtered_with_id.createOrReplaceTempView(src_view_name_filtered_with_id)
        
        # 5. end-date remaining matched current rows in target

        join_condition_tgt = " AND ".join([f"src.`tgt_{item}` = tgt.`{item}`" for item in key_columns])
        enddate_old_rows_sql = f"""
            MERGE INTO {db_reference}{self.__targetTableName} AS tgt
            USING {left_join_temp_name} AS src
            ON {join_condition_tgt}
            WHEN MATCHED AND tgt.`{current_row_column}` = TRUE THEN UPDATE {end_date_merge_update}
            """
        lgr.info('end-dating old rows')
        lgr.info(enddate_old_rows_sql)
        self._spark.sql(enddate_old_rows_sql)

        # 8. append all source rows to target
        
        lgr.info('appending new / updated data')

        writer = df_src_filtered_with_id.write
        if self.__partitionsDefined:
            writer = writer.partitionBy(self.__partitionByColumns)

        writer.mode("append").format("delta").save(self.__tableLocation)
        #df_src_filtered_with_id.write.mode("append").format("delta").save(self.__tableLocation)

        self._spark.sql(f'DROP TABLE IF EXISTS {src_view_name}')
        

    def __snapshotTarget(self):
        # first delete the snapshot to be replaced
        key_columns = list(map(lambda x: self.__sanitizeColumnName(x), self.__tableConfig.snapshotKeyColumns))
        
        if len(key_columns) == 0:
            raise Exception('No key columns set for snapshot increment. Please set the snapshotKeyColumns property in destinationTable configuration to the list of column names.')
        
        join_condition = " AND ".join([f"src.`{item}` = tgt.`{item}`" for item in key_columns])

        join_column_list = ", ".join([f"`{item}`" for item in key_columns])
        
        src_view_name = f"src_{self.__lhMeta.lakehouseName}_{self.__targetTableName}"
        self.__data.createOrReplaceTempView(src_view_name)
        mergeDbRef = f'{self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}'

        merge_delete_sql = f"MERGE INTO {mergeDbRef}{self.__targetTableName} AS tgt \
            USING (SELECT DISTINCT {join_column_list} FROM {src_view_name})  AS src \
            ON {join_condition} \
            WHEN MATCHED THEN DELETE \
            "
        
        self.__logger.info("----SNAPSHOT DELETE SQL")
        self.__logger.info(merge_delete_sql)
        #print(merge_delete_sql)
        self._spark.sql(merge_delete_sql)
    
        # append the new rows
        self.__appendTarget()

    def __tableExistsF(self):
        """Checks if a table in the lakehouse exists.
        """
        
        default_location = self.__lhBasePath + "/Tables/" + self.__targetTableName
        lowercase_location = self.__lhBasePath + "/Tables/" + self.__targetTableName.lower()
        if self.__useSchema:
            default_location = self.__lhBasePath + "/Tables/" + self.__targetSchemaName + '/' + self.__targetTableName
            lowercase_location = self.__lhBasePath + "/Tables/" + self.__targetSchemaName + '/' + self.__targetTableName.lower()
            
        self.__logger.info('checking the existence of ' + lowercase_location + ':')
        fileExists = notebookutils.mssparkutils.fs.exists(default_location) or notebookutils.mssparkutils.fs.exists(lowercase_location)
        self.__logger.info(fileExists)
        return fileExists
        
        # df_tables = self._spark.sql('SHOW TABLES')
        # df_f = df_tables.filter(lower('tableName') == self.__targetTableName.lower())
        # if df_f.count() > 0:
        #     return True
        # return False

    def __ensureTableIsReady(self, tableName: str):
        """After a delta table is written to the Tables/ folder in a lakehouse, 
        it can take a few seconds before it is available in SparkSQL.
        This function waits for up to 28 seconds (4 + 8 + 16) checking if the table is SQL-ready.
        When it's ready, the function returns. If it isn't, an exception is thrown.

        table_name is either a simple name or [lakehouse.table_name]
        """

        table_name_wrapped = tu.wrapInBackticks(tableName)
        wait_time = 2
        total_wait = 0
        while wait_time <= 16:
            try:
                df = self._spark.sql(f'SELECT * FROM {table_name_wrapped} LIMIT 0')
                return
            except Exception as e:
                if wait_time == 16:
                    print(f'The table {table_name_wrapped} still not found after {total_wait} seconds')
                    raise e
            wait_time = wait_time * 2
            time.sleep(wait_time)
            total_wait = total_wait + wait_time

    def __filterByWatermark(self):
        if not(self.__tableExists):
            return
        elif self.__incrementMethod is None:
            return
        elif self.__incrementMethod == 'overwrite':
            return
        
        watermarkColumn = self.__tableConfig.watermarkColumn
        if watermarkColumn is None:
            return
        
        max_watermark_sql = f'SELECT MAX(`{watermarkColumn}`) FROM {self.__lhMeta.lakehouseName}.{self.__sechemaRefPrefix}{self.__targetTableName}'
        # print(max_watermark_sql)
        max_watermark = self._spark.sql(max_watermark_sql).collect()[0][0]
        # print(f'max watermark: {max_watermark}')

        # if the table is empty or watermark is 0, take all the data
        if max_watermark is None:
            return
        if not max_watermark:
            return
        
        filter = f'{watermarkColumn} > "{max_watermark}"'
        self.__data = self.__data.filter(filter)

    def __getShowString(self, df, n=20, truncate=True, vertical=False):
        if isinstance(truncate, bool) and truncate:
            return(df._jdf.showString(n, 20, vertical))
        else:
            return(df._jdf.showString(n, int(truncate), vertical))
        # print(filter)
        # print(self.__data.count())


from bifabrik.cfg.engine.ConfigContainer import ConfigContainer
from bifabrik.cfg.specific.JdbcConfiguration import JdbcConfiguration
from bifabrik.cfg.DataSourceConfigurationBase import DataSourceConfigurationBase

class JdbcSourceConfiguration(DataSourceConfigurationBase):

    def __init__(self):
        self.__jdbc = JdbcConfiguration()
        super().__init__()
        

    @property
    def jdbc(self) -> JdbcConfiguration:
        return self.__jdbc


from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class JdbcConfiguration(Configuration):
    """Configuration related to reading JDBC endpoints
    """
    def __init__(self):
        super().__init__()
        self.__jdbcUrl = None
        # super().createCfgProperty("delimiter", """From pandas:
        # Character or regex pattern to treat as the delimiter. If sep=None, the C engine cannot automatically detect the separator, but the Python parsing engine can, meaning the latter will be used and automatically detect the separator from only the first valid row of the file by Pythons builtin sniffer tool
        # """)        
        self.__jdbcUser = None
        self.__jdbcQuery = None
        self.__jdbcDbtable = None
        self.__jdbcPrepareQuery = None
        self.__jdbcDriver = None
        self.__jdbcPartitionColumn = None
        self.__jdbcLowerBound = None
        self.__jdbcUpperBound = None
        self.__jdbcNumPartitions = None
        self.__jdbcQueryTimeout = None
        self.__jdbcFetchsize = None
        self.__jdbcBatchsize = None
        self.__jdbcIsolationLevel = None
        self.__jdbcSessionInitStatement = None
        self.__jdbcTruncate = None
        self.__jdbcCreateTableOptions = None
        self.__jdbcCreateTableColumnTypes = None
        self.__jdbcCustomSchema = None
        self.__jdbcPushDownPredicate = None
        self.__jdbcPushDownAggregate = None
        self.__jdbcPushDownLimit = None
        self.__jdbcPushDownOffset = None
        self.__jdbcPushDownTableSample = None
        self.__jdbcKeytab = None
        self.__jdbcPrincipal = None
        self.__jdbcRefreshKrb5Config = None
        self.__jdbcConnectionProvider = None
        self.__jdbcPreferTimestampNTZ = None
        

    @CfgProperty
    def jdbcUrl(self) -> str:
        """From pandas:
        Character or regex pattern to treat as the delimiter. If sep=None, the C engine cannot automatically detect the separator, but the Python parsing engine can, meaning the latter will be used and automatically detect the separator from only the first valid row of the file by Pythons builtin sniffer tool
        """
        return self.__jdbcUrl
    @jdbcUrl.setter(key='jdbcUrl')
    def jdbcUrl(self, val):
        self.__jdbcUrl = val
            
    @CfgProperty
    def jdbcUser(self) -> str:
        """The JDBC user name to connect with."""
        return self.__jdbcUser
    @jdbcUser.setter(key='jdbcUser')
    def jdbcUser(self, val):
        self.__jdbcUser = val

    @CfgProperty
    def jdbcQuery(self) -> str:
        """The JDBC query to execute."""
        return self.__jdbcQuery
    @jdbcQuery.setter(key='jdbcQuery')
    def jdbcQuery(self, val):
        self.__jdbcQuery = val

    @CfgProperty
    def jdbcDbtable(self) -> str:
        """The JDBC database table to connect to."""
        return self.__jdbcDbtable
    @jdbcDbtable.setter(key='jdbcDbtable')
    def jdbcDbtable(self, val):
        self.__jdbcDbtable = val

    @CfgProperty
    def jdbcPrepareQuery(self) -> str:
        """The JDBC prepared query to execute."""
        return self.__jdbcPrepareQuery
    @jdbcPrepareQuery.setter(key='jdbcPrepareQuery')
    def jdbcPrepareQuery(self, val):
        self.__jdbcPrepareQuery = val

    @CfgProperty
    def jdbcDriver(self) -> str:
        """The JDBC driver to use."""
        return self.__jdbcDriver
    @jdbcDriver.setter(key='jdbcDriver')
    def jdbcDriver(self, val):
        self.__jdbcDriver = val

    @CfgProperty
    def jdbcPartitionColumn(self) -> str:
        """The JDBC partition column."""
        return self.__jdbcPartitionColumn
    @jdbcPartitionColumn.setter(key='jdbcPartitionColumn')
    def jdbcPartitionColumn(self, val):
        self.__jdbcPartitionColumn = val

    @CfgProperty
    def jdbcLowerBound(self) -> int:
        """The JDBC lower bound for partitioning."""
        return self.__jdbcLowerBound
    @jdbcLowerBound.setter(key='jdbcLowerBound')
    def jdbcLowerBound(self, val):
        self.__jdbcLowerBound = val

    @CfgProperty
    def jdbcUpperBound(self) -> int:
        """The JDBC upper bound for partitioning."""
        return self.__jdbcUpperBound
    @jdbcUpperBound.setter(key='jdbcUpperBound')
    def jdbcUpperBound(self, val):
        self.__jdbcUpperBound = val

    @CfgProperty
    def jdbcNumPartitions(self) -> int:
        """The number of JDBC partitions."""
        return self.__jdbcNumPartitions
    @jdbcNumPartitions.setter(key='jdbcNumPartitions')
    def jdbcNumPartitions(self, val):
        self.__jdbcNumPartitions = val

    @CfgProperty
    def jdbcQueryTimeout(self) -> int:
        """The JDBC query timeout in seconds."""
        return self.__jdbcQueryTimeout
    @jdbcQueryTimeout.setter(key='jdbcQueryTimeout')
    def jdbcQueryTimeout(self, val):
        self.__jdbcQueryTimeout = val

    @CfgProperty
    def jdbcFetchsize(self) -> int:
        """The JDBC fetch size."""
        return self.__jdbcFetchsize
    @jdbcFetchsize.setter(key='jdbcFetchsize')
    def jdbcFetchsize(self, val):
        self.__jdbcFetchsize = val

    @CfgProperty
    def jdbcBatchsize(self) -> int:
        """The JDBC batch size."""
        return self.__jdbcBatchsize
    @jdbcBatchsize.setter(key='jdbcBatchsize')
    def jdbcBatchsize(self, val):
        self.__jdbcBatchsize = val

    @CfgProperty
    def jdbcIsolationLevel(self) -> str:
        """The JDBC isolation level."""
        return self.__jdbcIsolationLevel
    @jdbcIsolationLevel.setter(key='jdbcIsolationLevel')
    def jdbcIsolationLevel(self, val):
        self.__jdbcIsolationLevel = val

    @CfgProperty
    def jdbcSessionInitStatement(self) -> str:
        """The JDBC session initialization statement."""
        return self.__jdbcSessionInitStatement
    @jdbcSessionInitStatement.setter(key='jdbcSessionInitStatement')
    def jdbcSessionInitStatement(self, val):
        self.__jdbcSessionInitStatement = val

    @CfgProperty
    def jdbcTruncate(self) -> bool:
        """Whether to truncate the JDBC table before loading."""
        return self.__jdbcTruncate
    @jdbcTruncate.setter(key='jdbcTruncate')
    def jdbcTruncate(self, val):
        self.__jdbcTruncate = val

    @CfgProperty
    def jdbcCreateTableOptions(self) -> str:
        """The JDBC create table options."""
        return self.__jdbcCreateTableOptions
    @jdbcCreateTableOptions.setter(key='jdbcCreateTableOptions')
    def jdbcCreateTableOptions(self, val):
        self.__jdbcCreateTableOptions = val

    @CfgProperty
    def jdbcCreateTableColumnTypes(self) -> str:
        """The JDBC create table column types."""
        return self.__jdbcCreateTableColumnTypes
    @jdbcCreateTableColumnTypes.setter(key='jdbcCreateTableColumnTypes')
    def jdbcCreateTableColumnTypes(self, val):
        self.__jdbcCreateTableColumnTypes = val

    @CfgProperty
    def jdbcCustomSchema(self) -> str:
        """The JDBC custom schema."""
        return self.__jdbcCustomSchema
    @jdbcCustomSchema.setter(key='jdbcCustomSchema')
    def jdbcCustomSchema(self, val):
        self.__jdbcCustomSchema = val

    @CfgProperty
    def jdbcPushDownPredicate(self) -> bool:
        """Whether to push down predicates to the JDBC source."""
        return self.__jdbcPushDownPredicate
    @jdbcPushDownPredicate.setter(key='jdbcPushDownPredicate')
    def jdbcPushDownPredicate(self, val):
        self.__jdbcPushDownPredicate = val

    @CfgProperty
    def jdbcPushDownAggregate(self) -> bool:
        """Whether to push down aggregates to the JDBC source."""
        return self.__jdbcPushDownAggregate
    @jdbcPushDownAggregate.setter(key='jdbcPushDownAggregate')
    def jdbcPushDownAggregate(self, val):
        self.__jdbcPushDownAggregate = val

    @CfgProperty
    def jdbcPushDownLimit(self) -> int:
        """The JDBC push down limit."""
        return self.__jdbcPushDownLimit
    @jdbcPushDownLimit.setter(key='jdbcPushDownLimit')
    def jdbcPushDownLimit(self, val):
        self.__jdbcPushDownLimit = val

    @CfgProperty
    def jdbcPushDownOffset(self) -> int:
        """The JDBC push down offset."""
        return self.__jdbcPushDownOffset
    @jdbcPushDownOffset.setter(key='jdbcPushDownOffset')
    def jdbcPushDownOffset(self, val):
        self.__jdbcPushDownOffset = val

    @CfgProperty
    def jdbcPushDownTableSample(self) -> float:
        """The JDBC push down table sample percentage."""
        return self.__jdbcPushDownTableSample
    @jdbcPushDownTableSample.setter(key='jdbcPushDownTableSample')
    def jdbcPushDownTableSample(self, val):
        self.__jdbcPushDownTableSample = val

    @CfgProperty
    def jdbcKeytab(self) -> str:
        """The JDBC keytab file for Kerberos authentication."""
        return self.__jdbcKeytab
    @jdbcKeytab.setter(key='jdbcKeytab')
    def jdbcKeytab(self, val):
        self.__jdbcKeytab = val

    @CfgProperty
    def jdbcPrincipal(self) -> str:
        """The JDBC principal for Kerberos authentication."""
        return self.__jdbcPrincipal
    @jdbcPrincipal.setter(key='jdbcPrincipal')
    def jdbcPrincipal(self, val):
        self.__jdbcPrincipal = val

    @CfgProperty
    def jdbcRefreshKrb5Config(self) -> bool:
        """Whether to refresh the Kerberos configuration."""
        return self.__jdbcRefreshKrb5Config
    @jdbcRefreshKrb5Config.setter(key='jdbcRefreshKrb5Config')
    def jdbcRefreshKrb5Config(self, val):
        self.__jdbcRefreshKrb5Config = val

    @CfgProperty
    def jdbcConnectionProvider(self) -> str:
        """The JDBC connection provider."""
        return self.__jdbcConnectionProvider
    @jdbcConnectionProvider.setter(key='jdbcConnectionProvider')
    def jdbcConnectionProvider(self, val):
        self.__jdbcConnectionProvider = val

    @CfgProperty
    def jdbcPreferTimestampNTZ(self) -> bool:
        """Whether to prefer timestamp without time zone."""
        return self.__jdbcPreferTimestampNTZ
    @jdbcPreferTimestampNTZ.setter(key='jdbcPreferTimestampNTZ')
    def jdbcPreferTimestampNTZ(self, val):
        self.__jdbcPreferTimestampNTZ = val



from bifabrik.cfg.engine.Configuration import Configuration
from bifabrik.cfg.engine.Configuration import CfgProperty

class TableConfiguration(Configuration):
    """Configuration related to saving data to delta tables in a Fabric lakehouse or warehouse
    """
    def __init__(self):
        super().__init__()
        self.__watermarkColumn = None
        self.__increment = None
        self.__identityColumnPattern = None
        self.__insertDateColumn = None
        self.__updateDateColumn = None
        self.__mergeKeyColumns = []
        self.__snapshotKeyColumns = []
        self.__invalidCharactersInColumnNamesReplacement = '_'
        self.__canAddNewColumns = True
        self.__allowMissingColumnsInSource = True
        self.__addNARecord = False
        self.__addBadValueRecord = False
        self.__largeTableMergeMethodEnabled = False
        self.__largeTableMergeMethodSourceThresholdGB: float = 0.2
        self.__largeTableMergeMethodDestinationThresholdGB: float = 20
        self.__mergeSourceMaterializationThresholdRowcount: int = 1000000
        self.__rowStartColumn = 'RowStart'
        self.__rowEndColumn = 'RowEnd'
        self.__currentRowColumn = 'CurrentRow'
        self.__scd2SoftDelete = False
        self.__scd2ExcludeColumns = []
        self.__partitionByColumns = []
        self.__ensureTableIsReadyAfterSaving = True

    @CfgProperty
    def watermarkColumn(self) -> str:
        """Which column to use as the watermark when filtering records for incremental load.
        If the target table already exists and the watermark column is set, only records where the watermark is greater than the MAX from the table will be inserted.
        default None
        """
        return self.__watermarkColumn
    @watermarkColumn.setter(key='watermarkColumn')
    def watermarkColumn(self, val):
        self.__watermarkColumn = val

    @CfgProperty
    def increment(self) -> str:
        """Increment method
        None or "overwrite" : full load
        "append"            : add incoming rows to the table
        "merge"             : update / insert (SCD 1) after matching based on the mergeKeyColumns setting
        "scd2"              : perform SCD2 matching rows based on the mergeKeyColumns setting, using RowStartColumn, RowEndColumn and CurrentRowColumn column names 
        "snapshot"          : based on the snapshotKeyColumns setting (typically a date), inseert the new records and remove any old records that refer to a newly loaded snapshot key
        default None
        """
        return self.__increment
    @increment.setter(key='increment')
    def increment(self, val):
        if val is not None:
            val = val.lower()
        self.__increment = val

    @CfgProperty
    def mergeKeyColumns(self) -> str:
        """Array of column names used as the business key columns when the 'merge' increment method is used
        default []
        """
        return self.__mergeKeyColumns
    @mergeKeyColumns.setter(key='mergeKeyColumns')
    def mergeKeyColumns(self, val):
        t = str(type(val))
        if t == "<class 'str'>":
            self.__mergeKeyColumns = val.split(',')
        else:
            self.__mergeKeyColumns = val

    @CfgProperty
    def snapshotKeyColumns(self) -> str:
        """Array of column names used as the snapshot key columns when the 'snapshot' increment method is used
        default []
        """
        return self.__snapshotKeyColumns
    @snapshotKeyColumns.setter(key='snapshotKeyColumns')
    def snapshotKeyColumns(self, val):
        self.__snapshotKeyColumns = val

    @CfgProperty
    def identityColumnPattern(self) -> str:
        """The name pattern of the identity column to be added to the table. The values are 1, 2, 3..., each new record gets a new number.
        If configured, the column will be placed at the beginning of the table.
        
        Supported placeholders:
        {tablename} : the delta table name
        {lakehousename} : the lakehouse name

        example: '{tablename}_ID'

        default None
        """
        return self.__identityColumnPattern
    @identityColumnPattern.setter(key='identityColumnPattern')
    def identityColumnPattern(self, val):
        self.__identityColumnPattern = val

    @CfgProperty
    def insertDateColumn(self) -> str:
        """The name of the timestamp column to be added to each record, reflecting the time when the record was inserted.
        default None
        """
        return self.__insertDateColumn
    @insertDateColumn.setter(key='insertDateColumn')
    def insertDateColumn(self, val):
        self.__insertDateColumn = val
    
    @CfgProperty
    def updateDateColumn(self) -> str:
        """The name of the timestamp column to be added to each record, reflecting the time when the record was inserted or last updated.
        This gets updated during SCD1 updates, unlike insertDateColumn
        default None
        """
        return self.__updateDateColumn
    @updateDateColumn.setter(key='updateDateColumn')
    def updateDateColumn(self, val):
        self.__updateDateColumn = val

    @CfgProperty
    def invalidCharactersInColumnNamesReplacement(self) -> str:
        """The characters ' ,;{}()\n\t=' are invalid in delta tables. This sets the replacement character. You can use '' to just remove the invalid characters.
        default '_'
        """
        return self.__invalidCharactersInColumnNamesReplacement
    @invalidCharactersInColumnNamesReplacement.setter(key='invalidCharactersInColumnNamesReplacement')
    def invalidCharactersInColumnNamesReplacement(self, val):
        self.__invalidCharactersInColumnNamesReplacement = val
    
    @CfgProperty
    def canAddNewColumns(self) -> bool:
        """If the table already exists and the input dataset has extra columns, add the columns to the table with empty values before loading the new input.
        default True
        """
        return self.__canAddNewColumns
    @canAddNewColumns.setter(key='canAddNewColumns')
    def canAddNewColumns(self, val):
        self.__canAddNewColumns = val

    @CfgProperty
    def allowMissingColumnsInSource(self) -> bool:
        """If the table already exists and the input dataset has missing columns, add the columns to the source with empty values.
        default True
        """
        return self.__allowMissingColumnsInSource
    @allowMissingColumnsInSource.setter(key='allowMissingColumnsInSource')
    def allowMissingColumnsInSource(self, val):
        self.__allowMissingColumnsInSource = val

    @CfgProperty
    def addNARecord(self) -> bool:
        """Add a dimension "N/A" record with identity -1, N/A for strings, 0 for numbers. identityColumnPattern needs to be configured when this is enabled.
        default False
        """
        return self.__addNARecord
    @addNARecord.setter(key='addNARecord')
    def addNARecord(self, val):
        self.__addNARecord = val

    @CfgProperty
    def addBadValueRecord(self) -> bool:
        """Add a dimension "Bad Value" record with identity 0, Bad Value for strings, 0 for numbers. identityColumnPattern needs to be configured when this is enabled.
        default False
        """
        return self.__addBadValueRecord
    @addBadValueRecord.setter(key='addBadValueRecord')
    def addBadValueRecord(self, val):
        self.__addBadValueRecord = val

    @CfgProperty
    def largeTableMergeMethodEnabled(self) -> bool:
        """For large tables in merge mode, first load the changed records to a separate temp table and perform the merge there before appending back to the original table
        default False
        """
        return self.__largeTableMergeMethodEnabled
    @largeTableMergeMethodEnabled.setter(key='largeTableMergeMethodEnabled')
    def largeTableMergeMethodEnabled(self, val):
        self.__largeTableMergeMethodEnabled = val

    @CfgProperty
    def largeTableMergeMethodSourceThresholdGB(self) -> float:
        """For large tables in merge mode, first load the changed records to a separate temp table and perform the merge there before appending back to the original table - threshold for the size of the source data in GB
        default 0.2
        """
        return self.__largeTableMergeMethodSourceThresholdGB
    @largeTableMergeMethodSourceThresholdGB.setter(key='largeTableMergeMethodSourceThresholdGB')
    def largeTableMergeMethodSourceThresholdGB(self, val):
        self.__largeTableMergeMethodSourceThresholdGB = val
    
    @CfgProperty
    def largeTableMergeMethodDestinationThresholdGB(self) -> float:
        """For large tables in merge mode, first load the changed records to a separate temp table and perform the merge there before appending back to the original table - threshold for the size of the destination table in GB
        default 20
        """
        return self.__largeTableMergeMethodDestinationThresholdGB
    @largeTableMergeMethodDestinationThresholdGB.setter(key='largeTableMergeMethodDestinationThresholdGB')
    def largeTableMergeMethodDestinationThresholdGB(self, val):
        self.__largeTableMergeMethodDestinationThresholdGB = val

    @CfgProperty
    def mergeSourceMaterializationThresholdRowcount(self) -> float:
        """For large tables, merge increment is more efficient if the source data is first materialized into a table and partitioned in the same manner as the target table to enable partition pruning.
        If the source rowcount is above this threshold, the table will be materialized before merging.
        default 1000000
        """
        return self.__mergeSourceMaterializationThresholdRowcount
    @mergeSourceMaterializationThresholdRowcount.setter(key='mergeSourceMaterializationThresholdRowcount')
    def mergeSourceMaterializationThresholdRowcount(self, val):
        self.__mergeSourceMaterializationThresholdRowcount = val

    @CfgProperty
    def rowStartColumn(self) -> str:
        """Only for SCD2 increment - the row start timestamp column name
        default RowStart
        """
        return self.__rowStartColumn
    @rowStartColumn.setter(key='rowStartColumn')
    def rowStartColumn(self, val):
        self.__rowStartColumn = val
    
    @CfgProperty
    def rowEndColumn(self) -> str:
        """Only for SCD2 increment - the row end timestamp column name
        default RowEnd
        """
        return self.__rowEndColumn
    @rowEndColumn.setter(key='rowEndColumn')
    def rowEndColumn(self, val):
        self.__rowEndColumn = val
    
    @CfgProperty
    def currentRowColumn(self) -> str:
        """Only for SCD2 increment - the current row (Ture / False) column name
        default CurrentRow
        """
        return self.__currentRowColumn
    @currentRowColumn.setter(key='currentRowColumn')
    def currentRowColumn(self, val):
        self.__currentRowColumn = val
    
    @CfgProperty
    def scd2SoftDelete(self) -> bool:
        """For SCD 2 increment, set the RowEnd timestamp for rows not found in source (for full load only)
        default False
        """
        return self.__scd2SoftDelete
    @scd2SoftDelete.setter(key='scd2SoftDelete')
    def scd2SoftDelete(self, val):
        self.__scd2SoftDelete = val

    @CfgProperty
    def scd2ExcludeColumns(self) -> bool:
        """For SCD 2 increment - a list of untracked non-key columns - these will not trigger a new row version if their value changes
        default []
        """
        return self.__scd2ExcludeColumns
    @scd2ExcludeColumns.setter(key='scd2ExcludeColumns')
    def scd2ExcludeColumns(self, val):
        t = str(type(val))
        if t == "<class 'str'>":
            self.__scd2ExcludeColumns = val.split(',')
        else:
            self.__scd2ExcludeColumns = val

    @CfgProperty
    def partitionByColumns(self) -> bool:
        """For SCD 2 increment - a list of columns by which to partition the destination table
        default []
        """
        return self.__partitionByColumns
    @partitionByColumns.setter(key='partitionByColumns')
    def partitionByColumns(self, val):
        t = str(type(val))
        if t == "<class 'str'>":
            self.__partitionByColumns = val.split(',')
        else:
            self.__partitionByColumns = val
    
    @CfgProperty
    def ensureTableIsReadyAfterSaving(self) -> bool:
        """It can take a few seconds before a newly created table is available in SparkSQL - with this option, the data saving pipeline waits for the table to be available
        default True
        """
        return self.__ensureTableIsReadyAfterSaving
    @ensureTableIsReadyAfterSaving.setter(key='ensureTableIsReadyAfterSaving')
    def ensureTableIsReadyAfterSaving(self, val):
        self.__ensureTableIsReadyAfterSaving = val


from bifabrik.src.DataSource import DataSource
from bifabrik.cfg.JdbcSourceConfiguration import JdbcSourceConfiguration
from pyspark.sql.dataframe import DataFrame
from bifabrik.utils import fsUtils
from bifabrik.utils import log
import notebookutils.mssparkutils.credentials

class JdbcSource(DataSource, JdbcSourceConfiguration):
    """Jdbc data source
    supports configuration options from https://spark.apache.org/docs/3.5.2/sql-data-sources-jdbc.html
    
    Examples
    --------
    > import bifabrik as bif
    >
    > bif.fromJdbc('SELECT * FROM Orders').url(';').user(',').passwordKVSecretName('passwordSecret').toTable('FactOrderLine').run()
    """
    
    def __init__(self, parentPipeline):
        super().__init__(parentPipeline)
        JdbcSourceConfiguration.__init__(self)
        #self.__processed_abfs_paths = []
        self.__mergedConfig = None

    def __str__(self):
        url = None
        if self.__mergedConfig is not None:
            if 'jdbcUrl' in self.__mergedConfig.jdbc._explicitProps:
                url = self.__mergedConfig.jdbc._explicitProps['jdbcUrl']
        return f'JDBC source: {url}'
    
    def execute(self, input):
        lgr = log.getLogger()
        mergedConfig = self._pipeline.configuration.mergeToCopy(self)
        self.__mergedConfig = mergedConfig

        # set spark options for all the configuration switches specified in the task's config
        readerBase = self._spark.read.format("jdbc")

        # get login and password from KV if provided
        loginKVSecretName = mergedConfig.security.loginKVSecretName
        passwordKVSecretName = mergedConfig.security.passwordKVSecretName
        if not (loginKVSecretName is None):
            username = notebookutils.mssparkutils.credentials.getSecret(mergedConfig.security.keyVaultUrl, loginKVSecretName)
            readerBase.option('user', username)
        if not (passwordKVSecretName is None):
            password = notebookutils.mssparkutils.credentials.getSecret(mergedConfig.security.keyVaultUrl, passwordKVSecretName)
            readerBase.option('password', password)

        # get JDBC-speific options in a loop
        # (if the jdbcUser option is specified, it overrides the user from loginKVSecretName)
        for key in mergedConfig.jdbc._explicitProps:
            sparkKey = key[4:]
            sparkKey = sparkKey[0].lower() + sparkKey[1:]
            
            lgr.info(f'Setting {sparkKey} to {mergedConfig.jdbc._explicitProps[key]}')
            #print(f'Setting {sparkKey} to {mergedConfig.jdbc._explicitProps[key]}')
            readerBase.option(sparkKey, mergedConfig.jdbc._explicitProps[key])

        df = readerBase.load()
        
        self._result = df
        self._completed = True

    def cleanup(self):
        pass


from bifabrik.utils.lineage.LineageExpressionId import LineageExpressionId
from bifabrik.utils.lineage.LineageDependency import LineageDependency
from bifabrik.utils.lineage.LineageTableColumnId import LineageTableColumnId
from typing import List
from typing import Set
from typing import Dict

"""
Add any dependencies discovered in this node to the list
Return list of sub-expressions
"""
def visit_node(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    res_expressions: list[LineageExpressionId] = []
    #try:
    class_name_1 = node.getClass().getName()
    class_name_2 = node.getClass().getSuperclass().getName()
    class_name_3 = node.getClass().getSuperclass().getSuperclass().getName()
    #class_name_4 = node.getClass().getSuperclass().getSuperclass().getSuperclass().getName()
    #class_name_5 = node.getClass().getSuperclass().getSuperclass().getSuperclass().getSuperclass().getName()

    
    # ###################
    # print('VISITING NODE')
    # print(class_name_1)
    # print(class_name_2)
    # print(class_name_3)
    # ###################


    #print(class_name_4)
    #print(class_name_5)
    
    # ####################
    # print(node)
    # ####################
    


    # optimistic expectations
    handled = True
    #print(class_name_1)
    # Match the day to predefined patterns
    match class_name_1:
        case "org.apache.spark.sql.catalyst.plans.logical.Project":
            return visit_project_node(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.Join":
            return visit_join_node(node, dependencies)
        case "org.apache.spark.sql.execution.datasources.LogicalRelation":
            return visit_logical_relation(node, dependencies)
        #case "org.apache.spark.sql.catalyst.expressions.Alias":
        #    return visit_alias_expression(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.UnaryExpression":
            return visit_unary_expression(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.Attribute":
            return visit_attribute(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.WithCTE":
            return visit_cte(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.CTERelationRef":
            return visit_cte_relation_ref(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.Union":
            return visit_union(node, dependencies)
        case "org.apache.spark.sql.execution.LogicalRDD":
            return visit_logical_rdd(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.LocalRelation":
            return visit_logical_rdd(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.OneRowRelation":
            return visit_logical_rdd(node, dependencies)
        case _:
            handled = False
            
    if handled:
        return []
    
    handled = True
    # these also appear in class_name_1 handler - not sure, but there may be some nodes that have these as the class name without any sublass specified
    match class_name_2:
        case "org.apache.spark.sql.catalyst.expressions.UnaryExpression":
            return visit_unary_expression(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.Attribute":
            return visit_attribute(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.BinaryOperator":
            return visit_binary_operator(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.Expression":
            return visit_expression(node, dependencies)
        case "org.apache.spark.sql.catalyst.plans.logical.AggregateLike":
            # GROUP BY
            return visit_aggregate(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.aggregate.AggregateFunction":
            return visit_aggregate_function(node, dependencies)
        # subqueries
        case "org.apache.spark.sql.catalyst.plans.QueryPlan":
            
            # print(node)
            # print(class_name_1)
            # print(class_name_2)
            # print(class_name_3)

            child_node = node.child()
            return visit_node(child_node, dependencies)
        case _:
            handled = False
    
    if handled:
        return {}

    handled = True
    match class_name_3:
        # subqueries
        case "org.apache.spark.sql.catalyst.plans.QueryPlan":
            
            # print(node)
            # print(class_name_1)
            # print(class_name_2)
            # print(class_name_3)

            child_node = node.child()
            return visit_node(child_node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.BinaryOperator":
            return visit_binary_operator(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.BinaryExpression":
            return visit_binary_operator(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.Expression":
            return visit_expression(node, dependencies)
        case "org.apache.spark.sql.catalyst.expressions.aggregate.AggregateFunction":
            return visit_aggregate_function(node, dependencies)
        case _:
            handled = False

    return res_expressions
    # ... return the IDs that were discovered while visiting this
    # except Exception as e:
    #     # TODO
    #     print('EXCEPTION')
    #     print(e)
    #     return res_expressions
    # ... return the IDs that were discovered while visiting this


"""Get the project items as a python list from the scala project node
"""
def get_project_list_python(project_list):
    class_name_1 = project_list.getClass().getName()
    class_name_2 = project_list.getClass().getSuperclass().getName()
    class_name_3 = project_list.getClass().getSuperclass().getSuperclass().getName()

    # print('PROJECT LIST WRAP')
    # print(class_name_1)
    # print(class_name_2)
    # print(class_name_3)

    # the project list can be either a list or a vector - get projection items in each case
    if class_name_1 == 'scala.collection.immutable.Vector':
        #print('PROJECT VECTOR')
        project_list_python_outer = [project_list.apply(i) for i in range(project_list.size())]
    elif class_name_3 == 'scala.collection.mutable.AbstractSeq':
        project_list_python_outer = [project_list.apply(i) for i in range(project_list.size())]
        
        # print('MUTABLE COLLECTION')
        # for x in project_list_python_outer:
        #     print(x)

        # scala.collection.mutable.ArrayBuffer
        # scala.collection.mutable.AbstractBuffer
        # scala.collection.mutable.AbstractSeq
    else:
        project_list_iterator = project_list.productIterator().toIndexedSeq()
        project_list_python_outer = [project_list_iterator.apply(i) for i in range(project_list_iterator.size())]
    

    #project_list_iterator = project_list.productIterator().toIndexedSeq()
    #project_list_python_outer = [project_list_iterator.apply(i) for i in range(project_list_iterator.size())]

    project_list_python = []
    for proj_outer in project_list_python_outer:
        proj_outer_class_name = proj_outer.getClass().getName()
        if proj_outer_class_name == "scala.collection.immutable.Nil$":
            continue
        if proj_outer_class_name == "scala.collection.immutable.$colon$colon":
            proj_elems = proj_outer.toIndexedSeq()
            proj_elems_python = [proj_elems.apply(i) for i in range(proj_elems.size())]
            for proj_elem in proj_elems_python:
                project_list_python.append(proj_elem)
        else:
            project_list_python.append(proj_outer)
    
    return project_list_python

def visit_project_node(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    
    #print('VISITING PROJECTION')
    
    project_list = node.projectList()
    res_expressions: list[LineageExpressionId] = []

    #print('PROJECT LIST')

    project_list_python = get_project_list_python(project_list)

    # print('PROJECT ITEMS')
    # for proj in project_list_python:
    #     print(proj)

    for proj in project_list_python:    
        proj_class_name = proj.getClass().getName()
        proj_class_name2 = proj.getClass().getSuperclass().getName()
        
        expr_id = LineageExpressionId(proj)
        

        #print(f'searching for dependencies of project item {expr_id}')
        sub_references = visit_node(proj, dependencies)
        for sub_ref in sub_references:
            dependencies.append(LineageDependency(source = sub_ref, target = expr_id))
            
            
            # dependent_node, depends_on_node, node_impact, node_dependencies
            #add_dependency(expr_id.id, sub_ref.id, )

        #res_references.add(expr_id.id)
        #res_references.update(sub_references)
        res_expressions.append(expr_id)
        res_expressions.extend(sub_references)
    
    proj_child = node.child()
    
    child_expressions = visit_node(proj_child, dependencies)
    res_expressions.extend(child_expressions)
    #res_references.update(child_refs)

    return res_expressions

def visit_join_node(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    #print('VISITING JOIN')

    res_expressions: list[LineageExpressionId] = []
    
    left = node.left()
    right = node.right()

    left_sub_ref = visit_node(left, dependencies)
    right_sub_ref = visit_node(right, dependencies)

    res_expressions.extend(left_sub_ref)
    res_expressions.extend(right_sub_ref)

    return res_expressions


def visit_logical_relation(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    #print('VISITING LOGICAL RELATION')
    
    res_expressions: list[LineageExpressionId] = []

    #relations.append(node)

    # print(r)

    # print(r.getClass().getName())
    # print(r.getClass().getSuperclass().getName())
    # print(r.getClass().getSuperclass().getSuperclass().getName())

    r = node
    
    # print('GETTING CATALOG TABLE FROM')
    # print(node)


    catalog_table = r.catalogTable()

    # print('---------------------')
    str_catalog_table = str(catalog_table)
    # print(str_catalog_table)
    # print('---------------------')

    table_identifier: str = None
    if str_catalog_table != 'None':
        table_identifier = r.catalogTable().get().identifier().toString()
        #print(table_identifier)

    

    r_pa = r.producedAttributes()
    r_pai = r_pa.toIndexedSeq()
    
    r_pal = [r_pai.apply(i) for i in range(r_pai.size())]

    for pa in r_pal:
        #print(pa)
        if table_identifier is not None:
            expr_id = LineageTableColumnId(pa, table_identifier)
        else:
            expr_id = LineageExpressionId(pa)
        res_expressions.append(expr_id)
        #print(expr_id)
        #print(pa.getClass().getName())


    return res_expressions

def visit_logical_rdd(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    # print('VISITING LOGICAL RDD')
    
    res_expressions: list[LineageExpressionId] = []

    r = node
    r_pa = r.producedAttributes()
    r_pai = r_pa.toIndexedSeq()
    
    r_pal = [r_pai.apply(i) for i in range(r_pai.size())]

    for pa in r_pal:
        expr_id = LineageExpressionId(pa)
        res_expressions.append(expr_id)
        # print(expr_id)
        # print(pa.getClass().getName())


    return res_expressions

def visit_cte_relation_ref(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    #print('VISITING CTE REF')
    
    res_expressions: list[LineageExpressionId] = []

    r = node

    r_pa = r.producedAttributes()
    r_pai = r_pa.toIndexedSeq()
    
    r_pal = [r_pai.apply(i) for i in range(r_pai.size())]

    # kind of redundant, these expressions appear in the CTE definition already
    for pa in r_pal:
        #print(pa)
        expr_id = LineageExpressionId(pa)
        res_expressions.append(expr_id)

    return res_expressions


def visit_unary_expression(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING UNARY EXPRESSION')
    
    #help(node)
    class_name_1 = node.getClass().getName()
    res_expressions: list[LineageExpressionId] = []
    
    ch = node.child()

    child_expressions = visit_node(ch, dependencies)
    res_expressions.extend(child_expressions)
    
    if class_name_1 == "org.apache.spark.sql.catalyst.expressions.Alias":
        #print('ALIAS')
        expression_id = LineageExpressionId(node) 
        res_expressions.append(expression_id)
        for ch_exp in child_expressions:
            dependencies.append(LineageDependency(source = ch_exp, target = expression_id))
        
        #aliases.append(node)

    return res_expressions


def visit_expression(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING GENERAL EXPRESSION')
    
    #help(node)
    class_name_1 = node.getClass().getName()
    res_expressions: list[LineageExpressionId] = []

    children = node.children().toIndexedSeq()
    children_python = [children.apply(i) for i in range(children.size())]
    
    for ch in children_python:
        child_expressions = visit_node(ch, dependencies)
        res_expressions.extend(child_expressions)

    return res_expressions


def visit_aggregate_function(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING AGGREGATE FUNCTION')
    
    #help(node)
    class_name_1 = node.getClass().getName()
    res_expressions: list[LineageExpressionId] = []
    
    #ch = node.child()

    children = node.children().toIndexedSeq()
    children_python = [children.apply(i) for i in range(children.size())]

    #print(f'children count: {len(children_python)}')
    for ch in children_python:
        child_expressions = visit_node(ch, dependencies)
        res_expressions.extend(child_expressions)
    
        #aliases.append(node)

    return res_expressions


def visit_attribute(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING ATTRIBUTE')
    
    #class_name_1 = node.getClass().getName()
    lineage_expression_id = LineageExpressionId(node)
    res_expressions: list[LineageExpressionId] = [lineage_expression_id]
    
    return res_expressions


def visit_binary_operator(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    # print('VISITING BINARY OPERATOR')
    
    #class_name_1 = node.getClass().getName()
    res_expressions: list[LineageExpressionId] = []
    
    l = node.left()
    r = node.right()

    # print('L')
    # print(l)
    # print('R')
    # print(r)

    child_expressions_left = visit_node(l, dependencies)
    child_expressions_right = visit_node(r, dependencies)
    
    res_expressions.extend(child_expressions_left)
    res_expressions.extend(child_expressions_right)
    
    #binary_operators.append(node)

    return res_expressions


def visit_aggregate(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING AGGREGATE')

    res_expressions: list[LineageExpressionId] = []

    a_seq = node.aggregateExpressions()
    aggregate_list_iterator = a_seq.toIndexedSeq()
    aggregate_list_python = [aggregate_list_iterator.apply(i) for i in range(aggregate_list_iterator.size())]

    for agg_item in aggregate_list_python:
        agg_class_name = agg_item.getClass().getName()
        agg_class_name2 = agg_item.getClass().getSuperclass().getName()

        expr_id = LineageExpressionId(agg_item)

        # print(f'searching for dependencies of aggregate item {expr_id}')
        # print('searching in')
        # print(agg_item)
        sub_references = visit_node(agg_item, dependencies)
        #print('sub_refs')
        for sub_ref in sub_references:
            #print(sub_ref)
            dependencies.append(LineageDependency(source = sub_ref, target = expr_id))

        res_expressions.append(expr_id)
        res_expressions.extend(sub_references)

    agg_child = node.child()

    child_expressions = visit_node(agg_child, dependencies)
    res_expressions.extend(child_expressions)


    #aggregates.append(node)

    return res_expressions

def visit_union(node, dependencies : list[LineageDependency]) -> list[LineageExpressionId]:
    #TODO: this is incomplete! only the first query in the UNION is mapped correctly
    
    # print('VISITING UNION')
    # print(node)
    class_name_1 = node.getClass().getName()
    #print(class_name_1)

    res_expressions: list[LineageExpressionId] = []

    children_python = [node]

    # convert from nested unions to a list of projections

    unwrapped = False
    while not unwrapped:
        unwrapped = True
        for i in range(0, len(children_python)):
            ch = children_python[i]
            child_type = ch.getClass().getName()
            if child_type == "org.apache.spark.sql.catalyst.plans.logical.Deduplicate":
                unwrapped = False
                # print('DE-DEDUPLICATE')
                # print(ch)
                ch_ch = ch.child()
                # print('DE-DEDUPLICATE CHILD')
                # print(ch_ch)
                children_python[i] = ch_ch
                break
            if child_type == "org.apache.spark.sql.catalyst.plans.logical.Union":
                #print('UNION IN UNION')
                unwrapped = False
                ch_children = ch.children().toIndexedSeq()
                ch_children_python = [ch_children.apply(i) for i in range(ch_children.size())]
                children_python[i:i+1] = ch_children_python
                #print('RECHECK')
                break

    # children = node.children().toIndexedSeq()
    # children_python = [children.apply(i) for i in range(children.size())]
    
# case "org.apache.spark.sql.catalyst.plans.logical.Union":
#             return visit_union(node, dependencies)

    union_output_columns = []


    # print(f'union children: {len(children_python)}')
    
    # for ch in children_python:
    #     print('UNION CHILD')
    #     print(ch)
    #     print(ch.getClass().getName())
    #     print(ch.getClass().getSuperclass().getName())

    first_child = True
    for ch in children_python:
        ch_class = ch.getClass().getName()
        ch_super_class = ch.getClass().getSuperclass().getName()
        
        # print('UNION CHILD')
        # print(ch)
        # print(ch_class)
        # print(ch_super_class)

        # UNION children can be projections or aggregates
        if ch_super_class == "org.apache.spark.sql.catalyst.plans.logical.AggregateLike":
            # print('AGGREGATE IN UNION')
            a_seq = ch.aggregateExpressions()
            aggregate_list_iterator = a_seq.toIndexedSeq()
            aggregate_list_python = [aggregate_list_iterator.apply(i) for i in range(aggregate_list_iterator.size())]
            project_list = aggregate_list_python
        elif ch_class == "org.apache.spark.sql.catalyst.plans.logical.Distinct":
            #help(ch)
            ch_ch = ch.child()
            ch_ch_class = ch_ch.getClass().getName()
            # print(ch_ch_class)
            project_list = get_project_list_python(ch_ch.projectList())
            
        # org.apache.spark.sql.catalyst.plans.logical.Distinct
        # org.apache.spark.sql.catalyst.plans.logical.LogicalPlan
        else: 
            project_list = get_project_list_python(ch.projectList())

        i = 0
        for proj in project_list:
            expr_id = LineageExpressionId(proj)
            
            sub_references = visit_node(proj, dependencies)
            for sub_ref in sub_references:
                #print('DEPENDENCY FROM UNION FIRST CHILD')
                dependencies.append(LineageDependency(source = sub_ref, target = expr_id))

            res_expressions.append(expr_id)
            res_expressions.extend(sub_references)

            
            # the first query in the union sets the output columns
            # all other queries are set as "inputs" of these output columns

            if first_child:
                union_output_columns.append(expr_id)
            else:
                dependencies.append(LineageDependency(source = expr_id, target = union_output_columns[i]))
            
            i = i + 1
            
        child_expressions = visit_node(ch, dependencies)
        res_expressions.extend(child_expressions)
        first_child = False


    return res_expressions

def visit_cte(node, dependencies : list[LineageDependency]) -> List[LineageExpressionId]:
    #print('VISITING CTE')
    
    #help(node)
    #class_name_1 = node.getClass().getName()
    res_expressions: list[LineageExpressionId] = []

    children = node.children().toIndexedSeq()
    children_python = [children.apply(i) for i in range(children.size())]
    
    for ch in children_python:
        child_expressions = visit_node(ch, dependencies)
        res_expressions.extend(child_expressions)

    return res_expressions